from __future__ import annotations

import inspect
from abc import ABC
from collections import defaultdict
from collections.abc import Iterable
from typing import Any, Callable, ClassVar, Literal, Optional, TypeVar

from cognite.client.data_classes.data_modeling import (
    ContainerApplyList,
    ContainerId,
    DataModelApply,
    DataModelId,
    EdgeApply,
    EdgeApplyList,
    InstancesApply,
    NodeApply,
    NodeApplyList,
    ViewApplyList,
    ViewId,
)
from pydantic import ValidationError
from pydantic.alias_generators import to_pascal, to_snake
from typing_extensions import Self

from cognite.powerops.client._generated.cogshop1.data_classes._core import DomainModelApply as DomainModelApplyCogShop1
from cognite.powerops.utils.serialization import get_pydantic_annotation

from .cdf_resources import CDFFile, CDFSequence
from .dms_models import PowerOpsDMSModel, PowerOpsDMSSourceModel
from .graph_ql import PowerOpsGraphQLModel
from .model import Model


class DataModel(Model, ABC):
    cls_by_container: ClassVar[dict[ContainerId, type[DomainModelApplyCogShop1]]]
    cls_by_view: ClassVar[dict[ViewId, type[DomainModelApplyCogShop1]]]
    graph_ql: ClassVar[Optional[PowerOpsGraphQLModel]] = None
    source_model: ClassVar[Optional[PowerOpsDMSSourceModel]] = None
    dms_model: ClassVar[Optional[PowerOpsDMSModel]] = None

    @classmethod
    def spaces(cls) -> list[str]:
        return list(
            {
                space
                for model in [cls.graph_ql, cls.source_model, cls.dms_model]
                if model is not None
                for space in model.spaces()
            }
        )

    @classmethod
    def data_model_ids(cls) -> list[DataModelId]:
        return list({model.id_ for model in [cls.graph_ql, cls.dms_model] if model is not None})

    @classmethod
    def containers(cls) -> ContainerApplyList | None:
        if not cls.source_model:
            return None
        return cls.source_model.containers or None

    @classmethod
    def views(cls) -> ViewApplyList | None:
        if not cls.dms_model:
            return None
        return cls.dms_model.views or None

    @classmethod
    def data_model(cls) -> DataModelApply | None:
        if not cls.dms_model:
            return None
        return cls.dms_model.data_model

    def instances(self) -> InstancesApply:
        nodes: dict[str, NodeApply] = {}
        edges: dict[str, EdgeApply] = {}
        for domain_model in self._domain_models():
            instance_applies = domain_model.to_instances_apply()
            # Caching in case recursive relationships are used.
            for node in instance_applies.nodes:
                if node.external_id not in nodes:
                    nodes[node.external_id] = node
            for edge in instance_applies.edges:
                if edge.external_id not in edges:
                    edges[edge.external_id] = edge

        return InstancesApply(nodes=NodeApplyList(nodes.values()), edges=EdgeApplyList(edges.values()))

    def nodes(self) -> NodeApplyList:
        return self.instances().nodes

    def edges(self) -> EdgeApplyList:
        return self.instances().edges

    cdf_resources: ClassVar[dict[Callable, type]] = {
        **dict(Model.cdf_resources.items()),
        nodes: NodeApply,
        edges: EdgeApply,
    }

    def _domain_models(self) -> Iterable[DomainModelApplyCogShop1]:
        for field_name in self.model_fields:
            items = getattr(self, field_name)
            if isinstance(items, list) and items and isinstance(items[0], DomainModelApplyCogShop1):
                yield from items
            if isinstance(items, DomainModelApplyCogShop1):
                yield items
            if isinstance(items, dict) and items and isinstance(next(iter(items.values())), DomainModelApplyCogShop1):
                yield from items.values()

    @classmethod
    def load_from_cdf_resources(
        cls: type[Self], data: dict[str, Any], link: Literal["external_id", "object"] = "object"
    ) -> Self:
        load_by_type_external_id = cls._load_by_type_external_id(data)
        if "nodes" not in load_by_type_external_id:
            return cls()
        nodes_by_source_by_id = defaultdict(dict)
        node_by_id = {}
        for node in load_by_type_external_id["nodes"].values():
            node: NodeApply
            if len(node.sources) != 1:
                raise ValueError(f"Node {node.external_id} has more than one source.")
            source = node.sources[0].source
            # Todo bug in SDK
            if isinstance(source, dict):
                if source.get("type") == "container":
                    source = ContainerId.load(source)
                elif source.get("type") == "view":
                    source = ViewId.load(source)
                else:
                    raise NotImplementedError("Cannot handle this source type.")
            if isinstance(source, ContainerId):
                if source not in cls.cls_by_container:
                    raise AttributeError(
                        f"Please specify the domain model for this container {source} in {cls.__name__}"
                    )

                domain_node = _load_domain_node(cls.cls_by_container[source], node)
            elif isinstance(source, ViewId):
                if source not in cls.cls_by_view:
                    raise AttributeError(f"Please specify the domain model for this view {source} in {cls.__name__}")
                domain_node = _load_domain_node(cls.cls_by_view[source], node)
            else:
                raise NotImplementedError(f"Cannot handle {type(source)} source type.")
            nodes_by_source_by_id[source.external_id][domain_node.external_id] = domain_node
            node_by_id[domain_node.external_id] = domain_node

        parsed = {}
        for field_name, field in cls.model_fields.items():
            annotation, outer = get_pydantic_annotation(field.annotation, cls)
            if issubclass(annotation, CDFFile) and outer is list:
                # Hack to set file names, in newer version files should always be linked to a type
                parsed[field_name] = list(load_by_type_external_id["files"].values())
            elif issubclass(annotation, CDFSequence) and outer is list:
                # Hack to set sequence names, in newer version sequences should always be linked to a type.
                suffix = field_name.removesuffix("s")
                parsed[field_name] = [
                    item for item in load_by_type_external_id["sequences"].values() if item.external_id.endswith(suffix)
                ]
            elif issubclass(annotation, DomainModelApplyCogShop1):
                name = field_name
                alternatives = [name, name.removesuffix("s"), to_pascal(name), to_pascal(name).removesuffix("s")]
                for alternative in alternatives:
                    if alternative in nodes_by_source_by_id:
                        name = alternative
                        break
                else:
                    # This means that there are no nodes for this field.
                    continue
                items = nodes_by_source_by_id[name]
                if outer is dict:
                    parsed[field_name] = dict(items)
                elif outer is list:
                    parsed[field_name] = list(items.values())
                else:
                    raise NotImplementedError()
            else:
                raise NotImplementedError()

        instance = cls(**parsed)

        instance_annotations = (
            get_pydantic_annotation(instance.model_fields[field_name].annotation, type(instance))[0]
            for field_name in instance.model_fields
        )
        model_types: set[type[DomainModelApplyCogShop1]] = {
            annotation for annotation in instance_annotations if issubclass(annotation, DomainModelApplyCogShop1)
        }

        # One to many edges
        edge_by_source_by_id = defaultdict(list)
        for edge in load_by_type_external_id.get("edges", {}).values():
            start_node = edge.start_node.external_id
            source_type, prop_name = edge.type.external_id.split(".", 1)
            edge_by_source_by_id[(start_node, source_type, to_snake(prop_name))].append(edge)

        for (source_id, source_type, prop_name), edges in edge_by_source_by_id.items():
            if source_type not in nodes_by_source_by_id or source_id not in nodes_by_source_by_id[source_type]:
                # Todo print warning
                # Missing source
                continue
            source = nodes_by_source_by_id[source_type][source_id]
            if prop_name not in source.model_fields:
                raise ValueError(f"Cannot find {prop_name} in {source}")
            annotation = source.model_fields[prop_name].annotation
            annotation, outer = get_pydantic_annotation(annotation, source)
            for edge in edges:
                if not (target := node_by_id.get(edge.end_node.external_id)):
                    # Todo print warning
                    # Missing target
                    continue
                if outer is list:
                    if getattr(source, prop_name) is None:
                        setattr(source, prop_name, [])
                    if link == "external_id" and annotation in model_types:
                        getattr(source, prop_name).append(target.external_id)
                    else:
                        getattr(source, prop_name).append(target)
                else:
                    raise NotImplementedError()

        # One to one edge
        for domain_node in node_by_id.values():
            for field_name, field in domain_node.model_fields.items():
                annotation, outer = get_pydantic_annotation(field.annotation, domain_node)
                if (
                    inspect.isclass(annotation)
                    and issubclass(annotation, DomainModelApplyCogShop1)
                    and (value := getattr(domain_node, field_name)) is not None
                ):
                    if isinstance(value, str) and value in node_by_id:
                        if link == "external_id" and annotation in model_types:
                            setattr(domain_node, field_name, value)
                        else:
                            setattr(domain_node, field_name, node_by_id[value])

        return instance


T_Domain_model = TypeVar("T_Domain_model", bound=DomainModelApplyCogShop1)


def _load_domain_node(node_cls: type[T_Domain_model], node: NodeApply) -> T_Domain_model:
    properties = node.sources[0].properties
    loaded = {}
    for name, prop in properties.items():
        snake_name = to_snake(name)
        if isinstance(prop, dict) and "externalId" in prop:
            loaded[snake_name] = prop["externalId"]
        elif isinstance(prop, (float, str, int)) or prop is None:
            loaded[snake_name] = prop
        elif isinstance(prop, list) and all(isinstance(p, (float, str, int)) for p in prop):
            loaded[snake_name] = prop
        elif isinstance(prop, dict):
            loaded[snake_name] = prop
        else:
            raise NotImplementedError(f"Cannot handle {prop=}")

    try:
        return node_cls(**loaded, external_id=node.external_id)
    except ValidationError:
        raise



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._alert import Alert, AlertGraphQL, AlertWrite
    from ._bid_method import BidMethod, BidMethodGraphQL, BidMethodWrite


__all__ = [
    "BidMatrix",
    "BidMatrixWrite",
    "BidMatrixApply",
    "BidMatrixList",
    "BidMatrixWriteList",
    "BidMatrixApplyList",
    "BidMatrixFields",
    "BidMatrixTextFields",
]


BidMatrixTextFields = Literal["resource_cost", "matrix", "asset_type", "asset_id"]
BidMatrixFields = Literal["resource_cost", "matrix", "asset_type", "asset_id"]

_BIDMATRIX_PROPERTIES_BY_FIELD = {
    "resource_cost": "resourceCost",
    "matrix": "matrix",
    "asset_type": "assetType",
    "asset_id": "assetId",
}


class BidMatrixGraphQL(GraphQLCore):
    """This represents the reading version of bid matrix, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid matrix.
        data_record: The data record of the bid matrix node.
        resource_cost: The resource cost field.
        matrix: The matrix field.
        asset_type: The asset type field.
        asset_id: The asset id field.
        method: The method field.
        alerts: The alert field.
    """

    view_id = dm.ViewId("power-ops-day-ahead-bid", "BidMatrix", "1")
    resource_cost: Optional[str] = Field(None, alias="resourceCost")
    matrix: Union[dict, None] = None
    asset_type: Optional[str] = Field(None, alias="assetType")
    asset_id: Optional[str] = Field(None, alias="assetId")
    method: Optional[BidMethodGraphQL] = Field(None, repr=False)
    alerts: Optional[list[AlertGraphQL]] = Field(default=None, repr=False)

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("method", "alerts", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> BidMatrix:
        """Convert this GraphQL format of bid matrix to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return BidMatrix(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            resource_cost=self.resource_cost,
            matrix=self.matrix["externalId"] if self.matrix and "externalId" in self.matrix else None,
            asset_type=self.asset_type,
            asset_id=self.asset_id,
            method=self.method.as_read() if isinstance(self.method, GraphQLCore) else self.method,
            alerts=[alert.as_read() if isinstance(alert, GraphQLCore) else alert for alert in self.alerts or []],
        )

    def as_write(self) -> BidMatrixWrite:
        """Convert this GraphQL format of bid matrix to the writing format."""
        return BidMatrixWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            resource_cost=self.resource_cost,
            matrix=self.matrix["externalId"] if self.matrix and "externalId" in self.matrix else None,
            asset_type=self.asset_type,
            asset_id=self.asset_id,
            method=self.method.as_write() if isinstance(self.method, DomainModel) else self.method,
            alerts=[alert.as_write() if isinstance(alert, DomainModel) else alert for alert in self.alerts or []],
        )


class BidMatrix(DomainModel):
    """This represents the reading version of bid matrix.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid matrix.
        data_record: The data record of the bid matrix node.
        resource_cost: The resource cost field.
        matrix: The matrix field.
        asset_type: The asset type field.
        asset_id: The asset id field.
        method: The method field.
        alerts: The alert field.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    resource_cost: Optional[str] = Field(None, alias="resourceCost")
    matrix: Union[str, None] = None
    asset_type: Optional[str] = Field(None, alias="assetType")
    asset_id: Optional[str] = Field(None, alias="assetId")
    method: Union[BidMethod, str, dm.NodeId, None] = Field(None, repr=False)
    alerts: Union[list[Alert], list[str], list[dm.NodeId], None] = Field(default=None, repr=False)

    def as_write(self) -> BidMatrixWrite:
        """Convert this read version of bid matrix to the writing version."""
        return BidMatrixWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            resource_cost=self.resource_cost,
            matrix=self.matrix,
            asset_type=self.asset_type,
            asset_id=self.asset_id,
            method=self.method.as_write() if isinstance(self.method, DomainModel) else self.method,
            alerts=[alert.as_write() if isinstance(alert, DomainModel) else alert for alert in self.alerts or []],
        )

    def as_apply(self) -> BidMatrixWrite:
        """Convert this read version of bid matrix to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidMatrixWrite(DomainModelWrite):
    """This represents the writing version of bid matrix.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid matrix.
        data_record: The data record of the bid matrix node.
        resource_cost: The resource cost field.
        matrix: The matrix field.
        asset_type: The asset type field.
        asset_id: The asset id field.
        method: The method field.
        alerts: The alert field.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    resource_cost: Optional[str] = Field(None, alias="resourceCost")
    matrix: Union[str, None] = None
    asset_type: Optional[str] = Field(None, alias="assetType")
    asset_id: Optional[str] = Field(None, alias="assetId")
    method: Union[BidMethodWrite, str, dm.NodeId, None] = Field(None, repr=False)
    alerts: Union[list[AlertWrite], list[str], list[dm.NodeId], None] = Field(default=None, repr=False)

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(BidMatrix, dm.ViewId("power-ops-day-ahead-bid", "BidMatrix", "1"))

        properties: dict[str, Any] = {}

        if self.resource_cost is not None or write_none:
            properties["resourceCost"] = self.resource_cost

        if self.matrix is not None or write_none:
            properties["matrix"] = self.matrix

        if self.asset_type is not None or write_none:
            properties["assetType"] = self.asset_type

        if self.asset_id is not None or write_none:
            properties["assetId"] = self.asset_id

        if self.method is not None:
            properties["method"] = {
                "space": self.space if isinstance(self.method, str) else self.method.space,
                "externalId": self.method if isinstance(self.method, str) else self.method.external_id,
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power-ops-types", "calculationIssue")
        for alert in self.alerts or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=alert,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.method, DomainModelWrite):
            other_resources = self.method._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class BidMatrixApply(BidMatrixWrite):
    def __new__(cls, *args, **kwargs) -> BidMatrixApply:
        warnings.warn(
            "BidMatrixApply is deprecated and will be removed in v1.0. Use BidMatrixWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "BidMatrix.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class BidMatrixList(DomainModelList[BidMatrix]):
    """List of bid matrixes in the read version."""

    _INSTANCE = BidMatrix

    def as_write(self) -> BidMatrixWriteList:
        """Convert these read versions of bid matrix to the writing versions."""
        return BidMatrixWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> BidMatrixWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidMatrixWriteList(DomainModelWriteList[BidMatrixWrite]):
    """List of bid matrixes in the writing version."""

    _INSTANCE = BidMatrixWrite


class BidMatrixApplyList(BidMatrixWriteList): ...


def _create_bid_matrix_filter(
    view_id: dm.ViewId,
    resource_cost: str | list[str] | None = None,
    resource_cost_prefix: str | None = None,
    asset_type: str | list[str] | None = None,
    asset_type_prefix: str | None = None,
    asset_id: str | list[str] | None = None,
    asset_id_prefix: str | None = None,
    method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(resource_cost, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("resourceCost"), value=resource_cost))
    if resource_cost and isinstance(resource_cost, list):
        filters.append(dm.filters.In(view_id.as_property_ref("resourceCost"), values=resource_cost))
    if resource_cost_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("resourceCost"), value=resource_cost_prefix))
    if isinstance(asset_type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("assetType"), value=asset_type))
    if asset_type and isinstance(asset_type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("assetType"), values=asset_type))
    if asset_type_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("assetType"), value=asset_type_prefix))
    if isinstance(asset_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("assetId"), value=asset_id))
    if asset_id and isinstance(asset_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("assetId"), values=asset_id))
    if asset_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("assetId"), value=asset_id_prefix))
    if method and isinstance(method, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("method"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": method}
            )
        )
    if method and isinstance(method, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("method"), value={"space": method[0], "externalId": method[1]})
        )
    if method and isinstance(method, list) and isinstance(method[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("method"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in method],
            )
        )
    if method and isinstance(method, list) and isinstance(method[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("method"), values=[{"space": item[0], "externalId": item[1]} for item in method]
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from typing import ClassVar, Optional

from cognite.client.data_classes import Asset
from pydantic import field_validator
from pydantic.dataclasses import Field

from cognite.powerops.resync.models.base import AssetModel, T_Asset_Type

from .base import Bid, DateTransformation, Market, Process, ShopTransformation
from .benchmark import BenchmarkBid, BenchmarkProcess, ProductionPlanTimeSeries
from .dayahead import DayAheadBid, DayAheadProcess, NordPoolMarket
from .rkom import RKOMBid, RKOMBidCombination, RKOMCombinationBid, RKOMMarket, RKOMPlants, RKOMProcess


class MarketModel(AssetModel):
    root_asset: ClassVar[Optional[Asset]] = None
    nordpool_market: list[NordPoolMarket] = Field(default_factory=list)
    rkom_market: list[RKOMMarket] = Field(default_factory=list)
    dayahead_processes: list[DayAheadProcess] = Field(default_factory=list)
    benchmark_processes: list[BenchmarkProcess] = Field(default_factory=list)
    rkom_processes: list[RKOMProcess] = Field(default_factory=list)
    combinations: list[RKOMBidCombination] = Field(default_factory=list)

    @field_validator(
        "nordpool_market",
        "rkom_market",
        "dayahead_processes",
        "benchmark_processes",
        "rkom_processes",
        "combinations",
        mode="after",
    )
    def ordering(cls, value: list[T_Asset_Type]) -> list[T_Asset_Type]:
        # To ensure loading the production model always yields the same result, we sort the assets by external_id.
        return sorted(value, key=lambda x: x.external_id)

    @classmethod
    def set_root_asset(
        cls, shop_service_url: str, organization_subdomain: str, tenant_id: str, core_root_asset_external_id: str
    ) -> None:
        if shop_service_url == "https://shop-staging.az-inso-powerops.cognite.ai/submit-run":
            customer = "cognite"
        else:
            customer = organization_subdomain

        cls.root_asset = Asset(
            parent_external_id=core_root_asset_external_id,
            external_id="configurations",
            name="Configurations",
            description="Configurations used for PowerOps",
            metadata={
                "shop_service_url": shop_service_url,
                "organization_subdomain": organization_subdomain,
                "customer": customer,
                "tenant_id": tenant_id,
            },
        )

    def standardize(self) -> None:
        self.dayahead_processes = self.ordering(self.dayahead_processes)
        self.benchmark_processes = self.ordering(self.benchmark_processes)
        self.rkom_processes = self.ordering(self.rkom_processes)
        self.combinations = self.ordering(self.combinations)
        for field in [
            self.dayahead_processes,
            self.benchmark_processes,
            self.rkom_processes,
            self.combinations,
            self.nordpool_market,
            self.rkom_market,
        ]:
            for item in field:
                item.standardize()

    @property
    def processes(self):
        return self.dayahead_processes + self.benchmark_processes + self.rkom_processes


__all__ = [
    "RKOMBid",
    "RKOMBidCombination",
    "RKOMMarket",
    "RKOMProcess",
    "RKOMCombinationBid",
    "RKOMPlants",
    "DayAheadBid",
    "NordPoolMarket",
    "DayAheadProcess",
    "BenchmarkBid",
    "BenchmarkProcess",
    "ProductionPlanTimeSeries",
    "Market",
    "DateTransformation",
    "ShopTransformation",
    "Bid",
    "Process",
    "MarketModel",
]



from ._core import (
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    GraphQLList,
    ResourcesWrite,
    ResourcesWriteResult,
    PageInfo,
)
from ._alert import (
    Alert,
    AlertApply,
    AlertApplyList,
    AlertFields,
    AlertGraphQL,
    AlertList,
    AlertTextFields,
    AlertWrite,
    AlertWriteList,
)
from ._basic_bid_matrix import (
    BasicBidMatrix,
    BasicBidMatrixApply,
    BasicBidMatrixApplyList,
    BasicBidMatrixFields,
    BasicBidMatrixGraphQL,
    BasicBidMatrixList,
    BasicBidMatrixTextFields,
    BasicBidMatrixWrite,
    BasicBidMatrixWriteList,
)
from ._bid_document import (
    BidDocument,
    BidDocumentApply,
    BidDocumentApplyList,
    BidDocumentFields,
    BidDocumentGraphQL,
    BidDocumentList,
    BidDocumentTextFields,
    BidDocumentWrite,
    BidDocumentWriteList,
)
from ._bid_matrix import (
    BidMatrix,
    BidMatrixApply,
    BidMatrixApplyList,
    BidMatrixFields,
    BidMatrixGraphQL,
    BidMatrixList,
    BidMatrixTextFields,
    BidMatrixWrite,
    BidMatrixWriteList,
)
from ._bid_method import (
    BidMethod,
    BidMethodApply,
    BidMethodApplyList,
    BidMethodFields,
    BidMethodGraphQL,
    BidMethodList,
    BidMethodTextFields,
    BidMethodWrite,
    BidMethodWriteList,
)
from ._multi_scenario_matrix import (
    MultiScenarioMatrix,
    MultiScenarioMatrixApply,
    MultiScenarioMatrixApplyList,
    MultiScenarioMatrixFields,
    MultiScenarioMatrixGraphQL,
    MultiScenarioMatrixList,
    MultiScenarioMatrixTextFields,
    MultiScenarioMatrixWrite,
    MultiScenarioMatrixWriteList,
)
from ._price_area import (
    PriceArea,
    PriceAreaApply,
    PriceAreaApplyList,
    PriceAreaFields,
    PriceAreaGraphQL,
    PriceAreaList,
    PriceAreaTextFields,
    PriceAreaWrite,
    PriceAreaWriteList,
)
from ._shop_multi_scenario_method import (
    SHOPMultiScenarioMethod,
    SHOPMultiScenarioMethodApply,
    SHOPMultiScenarioMethodApplyList,
    SHOPMultiScenarioMethodFields,
    SHOPMultiScenarioMethodGraphQL,
    SHOPMultiScenarioMethodList,
    SHOPMultiScenarioMethodTextFields,
    SHOPMultiScenarioMethodWrite,
    SHOPMultiScenarioMethodWriteList,
)
from ._shop_price_scenario import (
    SHOPPriceScenario,
    SHOPPriceScenarioApply,
    SHOPPriceScenarioApplyList,
    SHOPPriceScenarioFields,
    SHOPPriceScenarioGraphQL,
    SHOPPriceScenarioList,
    SHOPPriceScenarioTextFields,
    SHOPPriceScenarioWrite,
    SHOPPriceScenarioWriteList,
)
from ._shop_price_scenario_result import (
    SHOPPriceScenarioResult,
    SHOPPriceScenarioResultApply,
    SHOPPriceScenarioResultApplyList,
    SHOPPriceScenarioResultFields,
    SHOPPriceScenarioResultGraphQL,
    SHOPPriceScenarioResultList,
    SHOPPriceScenarioResultWrite,
    SHOPPriceScenarioResultWriteList,
)
from ._water_value_based_method import (
    WaterValueBasedMethod,
    WaterValueBasedMethodApply,
    WaterValueBasedMethodApplyList,
    WaterValueBasedMethodFields,
    WaterValueBasedMethodGraphQL,
    WaterValueBasedMethodList,
    WaterValueBasedMethodTextFields,
    WaterValueBasedMethodWrite,
    WaterValueBasedMethodWriteList,
)

BasicBidMatrix.model_rebuild()
BasicBidMatrixGraphQL.model_rebuild()
BasicBidMatrixWrite.model_rebuild()
BasicBidMatrixApply.model_rebuild()
BidDocument.model_rebuild()
BidDocumentGraphQL.model_rebuild()
BidDocumentWrite.model_rebuild()
BidDocumentApply.model_rebuild()
BidMatrix.model_rebuild()
BidMatrixGraphQL.model_rebuild()
BidMatrixWrite.model_rebuild()
BidMatrixApply.model_rebuild()
MultiScenarioMatrix.model_rebuild()
MultiScenarioMatrixGraphQL.model_rebuild()
MultiScenarioMatrixWrite.model_rebuild()
MultiScenarioMatrixApply.model_rebuild()
PriceArea.model_rebuild()
PriceAreaGraphQL.model_rebuild()
PriceAreaWrite.model_rebuild()
PriceAreaApply.model_rebuild()
SHOPMultiScenarioMethod.model_rebuild()
SHOPMultiScenarioMethodGraphQL.model_rebuild()
SHOPMultiScenarioMethodWrite.model_rebuild()
SHOPMultiScenarioMethodApply.model_rebuild()
SHOPPriceScenarioResult.model_rebuild()
SHOPPriceScenarioResultGraphQL.model_rebuild()
SHOPPriceScenarioResultWrite.model_rebuild()
SHOPPriceScenarioResultApply.model_rebuild()


__all__ = [
    "DataRecord",
    "DataRecordGraphQL",
    "DataRecordWrite",
    "ResourcesWrite",
    "DomainModel",
    "DomainModelCore",
    "DomainModelWrite",
    "DomainModelList",
    "DomainRelationWrite",
    "GraphQLCore",
    "GraphQLList",
    "ResourcesWriteResult",
    "PageInfo",
    "Alert",
    "AlertGraphQL",
    "AlertWrite",
    "AlertApply",
    "AlertList",
    "AlertWriteList",
    "AlertApplyList",
    "AlertFields",
    "AlertTextFields",
    "BasicBidMatrix",
    "BasicBidMatrixGraphQL",
    "BasicBidMatrixWrite",
    "BasicBidMatrixApply",
    "BasicBidMatrixList",
    "BasicBidMatrixWriteList",
    "BasicBidMatrixApplyList",
    "BasicBidMatrixFields",
    "BasicBidMatrixTextFields",
    "BidDocument",
    "BidDocumentGraphQL",
    "BidDocumentWrite",
    "BidDocumentApply",
    "BidDocumentList",
    "BidDocumentWriteList",
    "BidDocumentApplyList",
    "BidDocumentFields",
    "BidDocumentTextFields",
    "BidMatrix",
    "BidMatrixGraphQL",
    "BidMatrixWrite",
    "BidMatrixApply",
    "BidMatrixList",
    "BidMatrixWriteList",
    "BidMatrixApplyList",
    "BidMatrixFields",
    "BidMatrixTextFields",
    "BidMethod",
    "BidMethodGraphQL",
    "BidMethodWrite",
    "BidMethodApply",
    "BidMethodList",
    "BidMethodWriteList",
    "BidMethodApplyList",
    "BidMethodFields",
    "BidMethodTextFields",
    "MultiScenarioMatrix",
    "MultiScenarioMatrixGraphQL",
    "MultiScenarioMatrixWrite",
    "MultiScenarioMatrixApply",
    "MultiScenarioMatrixList",
    "MultiScenarioMatrixWriteList",
    "MultiScenarioMatrixApplyList",
    "MultiScenarioMatrixFields",
    "MultiScenarioMatrixTextFields",
    "PriceArea",
    "PriceAreaGraphQL",
    "PriceAreaWrite",
    "PriceAreaApply",
    "PriceAreaList",
    "PriceAreaWriteList",
    "PriceAreaApplyList",
    "PriceAreaFields",
    "PriceAreaTextFields",
    "SHOPMultiScenarioMethod",
    "SHOPMultiScenarioMethodGraphQL",
    "SHOPMultiScenarioMethodWrite",
    "SHOPMultiScenarioMethodApply",
    "SHOPMultiScenarioMethodList",
    "SHOPMultiScenarioMethodWriteList",
    "SHOPMultiScenarioMethodApplyList",
    "SHOPMultiScenarioMethodFields",
    "SHOPMultiScenarioMethodTextFields",
    "SHOPPriceScenario",
    "SHOPPriceScenarioGraphQL",
    "SHOPPriceScenarioWrite",
    "SHOPPriceScenarioApply",
    "SHOPPriceScenarioList",
    "SHOPPriceScenarioWriteList",
    "SHOPPriceScenarioApplyList",
    "SHOPPriceScenarioFields",
    "SHOPPriceScenarioTextFields",
    "SHOPPriceScenarioResult",
    "SHOPPriceScenarioResultGraphQL",
    "SHOPPriceScenarioResultWrite",
    "SHOPPriceScenarioResultApply",
    "SHOPPriceScenarioResultList",
    "SHOPPriceScenarioResultWriteList",
    "SHOPPriceScenarioResultApplyList",
    "SHOPPriceScenarioResultFields",
    "WaterValueBasedMethod",
    "WaterValueBasedMethodGraphQL",
    "WaterValueBasedMethodWrite",
    "WaterValueBasedMethodApply",
    "WaterValueBasedMethodList",
    "WaterValueBasedMethodWriteList",
    "WaterValueBasedMethodApplyList",
    "WaterValueBasedMethodFields",
    "WaterValueBasedMethodTextFields",
]



from __future__ import annotations

import difflib
import itertools
import logging
import re
import warnings
from collections import Counter, defaultdict
from dataclasses import asdict, dataclass
from graphlib import TopologicalSorter
from pathlib import Path
from typing import Any, Generic, TypeVar

import yaml
from cognite.client import CogniteClient
from cognite.client.data_classes import filters
from cognite.client.data_classes.data_modeling import (
    ContainerApply,
    ContainerApplyList,
    ContainerId,
    DataModelApply,
    DataModelApplyList,
    DataModelId,
    DirectRelation,
    MappedPropertyApply,
    NodeApply,
    NodeApplyList,
    NodeId,
    SpaceApply,
    SpaceApplyList,
    ViewApplyList,
    ViewId,
)
from cognite.client.data_classes.data_modeling.views import (
    MappedProperty,
    MultiEdgeConnection,
    SingleHopConnectionDefinitionApply,
    ViewApply,
    ViewList,
)

logger = logging.getLogger(__name__)

_DMS_DIR = Path(__file__).parent / "dms"


@dataclass
class Schema:
    _shared_space = "sp_powerops_models"
    containers: ContainerApplyList
    _views: ViewApplyList
    _data_models: DataModelApplyList
    spaces: SpaceApplyList
    node_types: NodeApplyList

    @property
    def container_views(self) -> ViewApplyList:
        # The data views are views that are a 1-1 mapping to a container
        # They are used to see what data is available in a container
        return ViewApplyList(
            [
                ViewApply(
                    space=self._shared_space,
                    external_id=f"{container.external_id}ContainerData_{container.space.replace('-','_')}",
                    version="1",
                    name=f"{container.name}ContainerData",
                    description=f"Data view for {container.name} in {container.space}",
                    properties={
                        prop: MappedPropertyApply(container=container.as_id(), container_property_identifier=prop)
                        for prop in container.properties
                    },
                )
                for container in self.containers
            ]
        )

    @property
    def container_model(self) -> DataModelApply:
        return DataModelApply(
            space=self._shared_space,
            external_id="PowerOpsContainerModel",
            version="1",
            name="PowerOpsContainers",
            description="All the data available in PowerOps",
            views=self.container_views.as_ids(),
        )

    @property
    def views(self) -> ViewApplyList:
        return self._views + self.container_views

    @property
    def data_models(self) -> DataModelApplyList:
        return self._data_models + DataModelApplyList([self.container_model])


@dataclass
class Difference:
    create: list[Any]
    changed: list[Any]
    unchanged: list[Any]
    delete: list[Any]

    def as_results(self, is_init: bool) -> list[dict]:
        if is_init:
            return (
                [{"action": "created", "model": item.as_id()} for item in self.create]
                + [{"action": "changed", "model": item.as_id()} for item in self.changed]
                + [{"action": "skipped-delete", "model": item.as_id()} for item in self.delete]
            )
        else:
            return [
                {"action": "deleted", "model": item.as_id()}
                for item in itertools.chain(self.delete, self.changed, self.unchanged)
            ]


T_ResourceList = TypeVar("T_ResourceList")


class DataModelAPI(Generic[T_ResourceList]):
    def __init__(self, class_api: Any, client: CogniteClient):
        self._client = client
        self._class_api = class_api

    @property
    def name(self) -> str:
        return type(self._class_api).__module__.rsplit(".", maxsplit=1)[-1]

    def apply(self, items: T_ResourceList) -> Any:
        return self._class_api.apply(items)

    def retrieve(self, spaces: SpaceApplyList) -> T_ResourceList:
        retrieved: T_ResourceList | None = None
        for space in spaces:
            if retrieved is None:
                retrieved = self._class_api.list(space=space.space)
            else:
                retrieved.extend(self._class_api.list(space=space.space))
        retrieved_apply = retrieved.as_apply() if retrieved else []
        if isinstance(retrieved_apply, ViewApplyList):
            # Need to remove properties that originate from other views
            view_by_id = {view.as_id(): view for view in retrieved_apply}
            for view in retrieved_apply:
                prop_to_pop: list[str] = []
                views_to_check = list(view.implements or [])
                while views_to_check:
                    view_to_check = views_to_check.pop()
                    prop_to_pop.extend(view_by_id[view_to_check].properties or [])
                    views_to_check.extend(view_by_id[view_to_check].implements or [])
                for prop in prop_to_pop:
                    view.properties.pop(prop, None)
        return retrieved_apply

    def differences(self, cdf: T_ResourceList, local: T_ResourceList) -> Difference:
        cdf_resources_by_id = {item.as_id(): item for item in cdf}
        local_resources_by_id = {item.as_id(): item for item in local}
        created = [item for item in local if item.as_id() not in cdf_resources_by_id]
        deleted = [item for item in cdf if item.as_id() not in local_resources_by_id]
        changed = []
        unchanged = []
        for item in local:
            if item.as_id() in cdf_resources_by_id:
                if cdf_resources_by_id[item.as_id()].dump() == item.dump():
                    unchanged.append(item)
                else:
                    changed.append(item)
        return Difference(create=created, changed=changed, unchanged=unchanged, delete=deleted)

    def delete(self, ids: list[Any]) -> None:
        return self._class_api.delete(ids)


class SpaceAPI(DataModelAPI[SpaceApplyList]):
    def retrieve(self, spaces: SpaceApplyList) -> SpaceApplyList:
        all_spaces = self._client.data_modeling.spaces.list()
        return SpaceApplyList([space.as_apply() for space in all_spaces if space.space.startswith("power-ops")])

    def delete(self, ids: list[Any]) -> None:
        for space in ids:
            if space == "power-ops-types":
                return  # Never delete the types space
            data_models = self._client.data_modeling.data_models.list(space=space)
            self._client.data_modeling.data_models.delete(data_models.as_ids())
            views = self._client.data_modeling.views.list(space=space)
            self._client.data_modeling.views.delete(views.as_ids())
            containers = self._client.data_modeling.containers.list(space=space)
            self._client.data_modeling.containers.delete(containers.as_ids())
            is_space = filters.Equals(["edge", "space"], space)
            edges = self._client.data_modeling.instances.list("edge", filter=is_space)
            self._client.data_modeling.instances.delete(edges.as_ids())
            is_space = filters.Equals(["node", "space"], space)
            nodes = self._client.data_modeling.instances.list(filter=is_space)
            self._client.data_modeling.instances.delete(nodes.as_ids())
            if nodes or edges or containers or views or data_models:
                print(
                    f"Deleted {space}: {len(nodes)} nodes, {len(edges)} edges, {len(containers)} containers, "
                    f"{len(views)} views, {len(data_models)} data models"
                )

        self._client.data_modeling.spaces.delete(ids)


class DataModelLoader:
    def __init__(self, source_dir: Path):
        self._source_dir = source_dir

    def load(self) -> Schema:
        resources_by_type = defaultdict(list)
        resource_cls_by_type = {
            "space": SpaceApply,
            "view": ViewApply,
            "container": ContainerApply,
            "node": NodeApply,
            "datamodel": DataModelApply,
        }

        failed: list[tuple[Path, str]] = []
        for filepath in self._source_dir.glob("**/*.yaml"):
            if filepath.name == "_build_environment.yaml":
                continue
            if match := re.match(r".*(?P<type>(space|view|container|powerops_nodes|datamodel))\.yaml$", filepath.name):
                type_ = match.group("type")
                if type_ == "powerops_nodes":
                    type_ = "node"
                resource_cls = resource_cls_by_type[type_]
                loaded = self._load_file(filepath)
                try:
                    if isinstance(loaded, list):
                        parsed = [resource_cls.load(item) for item in loaded]
                        resources_by_type[type_].extend(parsed)
                    else:
                        parsed = resource_cls.load(loaded)
                        resources_by_type[type_].append(parsed)
                except Exception as exc:
                    failed.append((filepath, str(exc)))
            else:
                raise ValueError(f"Unknown file type: {filepath}")
        if failed:
            raise ValueError(
                f"Failed to load {len(failed)} files:\n"
                + "\n".join(f"{filepath.relative_to(Path.cwd())}: {exc}" for filepath, exc in failed)
            )
        return Schema(
            containers=ContainerApplyList(resources_by_type["container"]),
            _views=ViewApplyList(resources_by_type["view"]),
            _data_models=DataModelApplyList(resources_by_type["datamodel"]),
            spaces=SpaceApplyList(resources_by_type["space"]),
            node_types=NodeApplyList(resources_by_type["node"]),
        )

    def _load_file(self, filepath: Path) -> dict[str, Any] | list[dict[str, Any]]:
        file_contents = filepath.read_text()
        return yaml.safe_load(file_contents)

    @classmethod
    def validate(cls, schema: Schema):
        defined_spaces = {space.space for space in schema.spaces}
        defined_containers = {container.as_id() for container in schema.containers}
        defined_node_types = {node_type.as_id() for node_type in schema.node_types}
        defined_views = {view.as_id() for view in schema._views}
        defined_interfaces = {parent for view in schema._views for parent in view.implements or []}
        properties_by_container = {container.as_id(): set(container.properties) for container in schema.containers}
        properties_by_view = {view.as_id(): set(view.properties) for view in schema._views}
        containers_by_id = {container.as_id(): container for container in schema.containers}

        referred_spaces = defaultdict(list)
        referred_views = defaultdict(list)
        referred_containers = defaultdict(list)
        referred_node_types = defaultdict(list)
        duplicated_views_by_model: dict[DataModelId, list[ViewId]] = {}
        double_referenced_container_properties: dict[ViewId, list] = defaultdict(list)
        view_missing_filters = []
        direct_relation_missing_source = []
        non_existent_container_properties = []
        non_existent_view_properties = []
        for container in schema.containers:
            referred_spaces[container.space].append(container.as_id())
        for view in schema._views:
            ref_view_id = view.as_id()
            referred_spaces[view.space].append(ref_view_id)
            if isinstance(view.filter, filters.And):
                filters_to_check = view.filter._filters
            else:
                filters_to_check = (view.filter,)
            for filter in filters_to_check:
                if isinstance(filter, filters.Equals):
                    dumped = filter.dump()["equals"]
                    value = dumped["value"]
                    try:
                        if isinstance(value, dict) and ("space" in value and "externalId" in value):
                            node_id = NodeId(value["space"], value["externalId"])
                            referred_node_types[node_id].append(ref_view_id)
                        elif len(dumped["property"]) == 3:
                            space, external_id_version, prop_ = dumped["property"]
                            if "/" in external_id_version:
                                # View Ref
                                external_id, version = external_id_version.rsplit("/", maxsplit=1)
                                view_id = ViewId(space, external_id, version)
                                referred_views[view_id].append(ref_view_id)
                                if prop_ not in properties_by_view.get(view_id, set()):
                                    if view.implements:
                                        for parent_view_id in view.implements:
                                            if prop_ not in properties_by_view.get(parent_view_id, set()):
                                                non_existent_view_properties.append(ref_view_id)
                                    else:
                                        non_existent_view_properties.append(ref_view_id)
                            else:
                                external_id = external_id_version
                                container_id = ContainerId(space, external_id)
                                referred_containers[container_id].append(ref_view_id)
                                if prop_ not in properties_by_container.get(container_id, set()):
                                    non_existent_container_properties.append(ref_view_id)

                    except Exception as exc:
                        raise ValueError(
                            f"Failed to parse filter for view {view.space}.{view.external_id}.{view.version}"
                        ) from exc
                elif isinstance(filter, filters.In):
                    dumped = filter.dump()["in"]["values"]
                    try:
                        for value in dumped:
                            if "space" in value and "externalId" in value:
                                node_id = NodeId(value["space"], value["externalId"])
                                referred_node_types[node_id].append(ref_view_id)
                    except Exception as exc:
                        raise ValueError(
                            f"Failed to parse filter for view {view.space}.{view.external_id}.{view.version}"
                        ) from exc

                if ref_view_id in defined_interfaces and not isinstance(filter, filters.In):
                    view_missing_filters.append(ref_view_id)
                elif ref_view_id not in defined_interfaces and isinstance(filter, filters.Equals):
                    view_missing_filters.append(ref_view_id)

            for view_id in view.implements or []:
                referred_views[view_id].append(ref_view_id)
            for prop_name, prop in view.properties.items():
                if isinstance(prop, MappedPropertyApply):
                    referred_containers[prop.container].append(ref_view_id.as_property_ref(prop_name))
                    if prop.source:
                        referred_views[prop.source].append(ref_view_id.as_property_ref(prop_name))
                    if prop.container_property_identifier not in properties_by_container.get(prop.container, set()):
                        non_existent_container_properties.append(ref_view_id.as_property_ref(prop_name))
                    if prop.container_property_identifier in properties_by_container.get(prop.container, set()):
                        container_property_relation_type = (
                            containers_by_id[prop.container].properties[prop.container_property_identifier].type
                        )
                        if type(container_property_relation_type) == DirectRelation and not prop.source:
                            direct_relation_missing_source.append(ref_view_id.as_property_ref(prop_name))
                elif isinstance(prop, SingleHopConnectionDefinitionApply):
                    referred_node_types[NodeId(prop.type.space, prop.type.external_id)].append(
                        ref_view_id.as_property_ref(prop_name)
                    )
                    referred_views[prop.source].append(ref_view_id.as_property_ref(prop_name))
                    if prop.edge_source:
                        referred_views[prop.edge_source].append(ref_view_id.as_property_ref(prop_name))

            counted = Counter(
                (prop.container, prop.container_property_identifier)
                for prop_name, prop in view.properties.items()
                if isinstance(prop, MappedPropertyApply)
            )
            for prop_identifier, count in counted.items():
                if count > 1:
                    double_referenced_container_properties[ref_view_id].append(prop_identifier)

        for node in schema.node_types:
            referred_spaces[node.space].append(node.as_id())
        for data_model in schema._data_models:
            referred_spaces[data_model.space].append(data_model.as_id())
            for view in data_model.views:
                referred_views[view].append(data_model.as_id())

            counted = Counter(view if isinstance(view, ViewId) else view.as_id() for view in data_model.views)
            duplicate_views = [view for view, count in counted.items() if count > 1]
            if duplicate_views:
                duplicated_views_by_model[data_model.as_id()] = duplicate_views

        if duplicated_views_by_model:
            message = "\n".join(f"{model_id}: {views}" for model_id, views in duplicated_views_by_model.items())
            raise ValueError(f"Duplicate views in data models:\n{message}")

        if undefined_spaces := set(referred_spaces).difference(defined_spaces):
            referred_to_by = list(itertools.chain(*(referred_spaces[space] for space in undefined_spaces)))
            raise ValueError(f"Undefined spaces: {undefined_spaces}: referred to by {referred_to_by}")
        if undefined_containers := set(referred_containers).difference(defined_containers):
            referred_to_by = list(
                itertools.chain(*(referred_containers[container] for container in undefined_containers))
            )
            raise ValueError(f"Undefined containers: {undefined_containers}: referred to by {referred_to_by}")
        if undefined_views := set(referred_views).difference(defined_views):
            referred_to_by = list(itertools.chain(*(referred_views[view] for view in undefined_views)))
            raise ValueError(f"Undefined views: {undefined_views}: referred to by {referred_to_by}")
        if undefined_node_types := set(referred_node_types).difference(defined_node_types):
            referred_to_by = list(
                itertools.chain(*(referred_node_types[node_type] for node_type in undefined_node_types))
            )
            raise ValueError(f"Undefined node types: {undefined_node_types}: referred to by {referred_to_by}")
        if non_existent_container_properties:
            message = "\n".join(map(str, non_existent_container_properties))
            raise ValueError(
                f"These properties in views refers to container properties that does not exist:\n{message}"
            )
        if non_existent_view_properties:
            message = "\n".join(map(str, non_existent_view_properties))
            raise ValueError(f"These properties in views refers to view properties that does not exist:\n{message}")
        if direct_relation_missing_source:
            message = "\n".join(map(str, direct_relation_missing_source))
            raise ValueError(f"These properties in views refers to direct relations without a source:\n{message}")

        if double_referenced_container_properties:
            message = "\n".join(
                f"{view_id}: {prop_identifier}"
                for view_id, prop_identifier in double_referenced_container_properties.items()
            )
            raise ValueError(f"These properties in views refers to container properties more than once:\n{message}")

    @classmethod
    def changed_views(cls, client: CogniteClient, schema: Schema, verbose: bool = False) -> dict[ViewId, set[ViewId]]:
        existing_read_views = client.data_modeling.views.retrieve(schema.views.as_ids())
        dependencies_by_view_id = cls._dependencies_by_view(existing_read_views)
        existing_views = cls._views_as_write(existing_read_views)

        existing_view_by_id = {view.as_id(): view for view in existing_views}
        changed_with_dependencies: dict[ViewId, set[ViewId]] = {}
        for view in schema.views:
            view_id = view.as_id()
            if (existing := existing_view_by_id.get(view_id)) and existing.dump() != view.dump():
                if verbose:
                    from rich import print
                    from rich.panel import Panel

                    # To get keys in the same order
                    new_view = existing.dump()
                    new_view.update(view.dump())
                    print(
                        Panel(
                            "\n".join(
                                difflib.unified_diff(
                                    existing.dump_yaml().splitlines(),
                                    yaml.safe_dump(new_view, sort_keys=False).splitlines(),
                                )
                            ),
                            title=f"Diff {view_id}",
                        )
                    )
                dependencies = dependencies_by_view_id[view_id]
                checked = set()
                while dependencies:
                    checked |= dependencies
                    dependencies = {
                        dep
                        for dep in itertools.chain(*(dependencies_by_view_id.get(dep, []) for dep in dependencies))
                        if dep not in checked
                    }
                changed_with_dependencies[view_id] = checked
        return changed_with_dependencies

    @classmethod
    def dependent_data_models(cls, schema: Schema, views: set[ViewId]) -> list[DataModelId]:
        dependent_data_models = []
        for data_model in schema.data_models:
            for view in data_model.views:
                view_id = view.as_id() if isinstance(view, ViewApply) else view
                if view_id in views:
                    dependent_data_models.append(data_model.as_id())
                    break
        return dependent_data_models

    @staticmethod
    def _views_as_write(views: ViewList) -> ViewApplyList:
        write_views = ViewApplyList([])
        view_by_id = {view.as_id(): view for view in views}
        for view in views:
            write_view = view.as_apply()
            for parent in view.implements or []:
                parent_properties = (parent_view := view_by_id.get(parent)) and parent_view.properties or {}
                for prop in parent_properties:
                    write_view.properties.pop(prop, None)
            write_views.append(write_view)
        return write_views

    @staticmethod
    def _dependencies_by_view(views: ViewApplyList):
        view_by_id = {view.as_id(): view for view in views}

        dependencies: dict[ViewId, set[ViewId]] = defaultdict(set)
        for view in views:
            view_id = view.as_id()
            for prop in view.properties.values():
                source: ViewId | None = None
                if isinstance(prop, MappedProperty) and isinstance(prop.type, DirectRelation) and prop.source:
                    source = prop.source
                elif isinstance(prop, MultiEdgeConnection):
                    source = prop.source

                if source and source in view_by_id:
                    dependencies[source].add(view_id)
                elif source:
                    warnings.warn(
                        f"The view {source} referenced by {view.as_id()} is not in the data model. Skipping it.",
                        stacklevel=2,
                    )

            for parent in view.implements or []:
                dependencies[parent].add(view_id)
            if view_id not in dependencies:
                dependencies[view_id] = set()
        return dependencies

    @classmethod
    def deploy(cls, client: CogniteClient, schema: Schema, is_dev: bool = False) -> list[dict]:
        result = []
        apis = cls._create_apis(client)
        resources = {
            "spaces": schema.spaces,
            "containers": schema.containers,
            "views": schema.views,
            "data_models": schema.data_models,
        }
        for api in TopologicalSorter(apis).static_order():
            items = resources[api.name]
            existing = api.retrieve(schema.spaces)
            diffs = api.differences(existing, items)
            if diffs.create or diffs.changed:
                if is_dev and api.name in {"views", "data_models"} and (diffs.changed or diffs.delete):
                    print(f"Deleting {api.name}: {len(diffs.changed + diffs.delete)} changed to avoid conflicts")
                    api.delete([item.as_id() for item in diffs.changed + diffs.delete])
                api.apply(diffs.create + diffs.changed)

            result.extend(diffs.as_results(is_init=True))
            print(
                f"Deployed {api.name}: {len(diffs.create)} created, {len(diffs.changed)} changed, "
                f"{len(diffs.unchanged)} unchanged, ({len(diffs.delete)} deleted - skipped)"
            )
        return result

    @classmethod
    def destroy(cls, client: CogniteClient, schema: Schema, dry_run: bool = False) -> list[dict]:
        result = []
        apis = cls._create_apis(client)
        resources = asdict(schema)
        for api in reversed(list(TopologicalSorter(apis).static_order())):
            try:
                items = resources[api.name]
            except KeyError:
                items = resources[f"_{api.name}"]
            existing = api.retrieve(schema.spaces)
            diffs = api.differences(existing, items)
            if not dry_run and (diffs.changed or diffs.unchanged or diffs.delete):
                api.delete(items.as_ids())
            total = len(diffs.changed + diffs.unchanged + diffs.delete)
            if dry_run:
                print(f"Would delete {api.name}: {total} deleted")
            else:
                print(f"Deleted {api.name}: {total} deleted")
            result.extend(diffs.as_results(is_init=False))

        return result

    @classmethod
    def _create_apis(cls, client: CogniteClient) -> dict[DataModelAPI, set[DataModelAPI]]:
        space = SpaceAPI(client.data_modeling.spaces, client)
        container = DataModelAPI[ContainerApplyList](client.data_modeling.containers, client)
        view = DataModelAPI[ViewApplyList](client.data_modeling.views, client)
        data_model = DataModelAPI[DataModelApplyList](client.data_modeling.data_models, client)
        return {space: set(), container: {space}, view: {space, container}, data_model: {space, container, view}}



from __future__ import annotations

import logging
import time
from enum import Enum
from typing import TYPE_CHECKING, Callable

from cognite.powerops.client.shop.data_classes.shop_run_event import ShopRunEvent

if TYPE_CHECKING:
    from cognite.powerops.client.shop.data_classes.shop_results import ShopRunResult

logger = logging.getLogger(__name__)


class ShopRun:
    """
    This represents a SHOP run.

    """

    class Status(Enum):
        IN_PROGRESS = "IN_PROGRESS"
        SUCCEEDED = "SUCCEEDED"
        FAILED = "FAILED"

    def __init__(
        self,
        check_status: Callable[[str], ShopRun.Status],
        retrieve_results: Callable[[ShopRun], ShopRunResult],
        *,
        shop_run_event: ShopRunEvent,
    ) -> None:
        self._check_status = check_status
        self._retrieve_results = retrieve_results
        self.shop_run_event = shop_run_event

    @property
    def in_progress(self) -> bool:
        """
        Returns whether the SHOP run is still running.

        Returns:
            True if the SHOP run is still running, False otherwise.
        """
        return self.status == ShopRun.Status.IN_PROGRESS

    @property
    def succeeded(self) -> bool:
        """
        Returns whether the SHOP run succeeded.

        Returns:
            True if the SHOP run succeeded, False otherwise.
        """
        return self.status == ShopRun.Status.SUCCEEDED

    @property
    def failed(self) -> bool:
        """
        Returns whether the SHOP run failed.

        Returns:
            True if the SHOP run failed, False otherwise.
        """
        return self.status == ShopRun.Status.FAILED

    @property
    def status(self) -> ShopRun.Status:
        """

        Queries the SHOP API for the status of the SHOP run.

        Returns:
            The status of the SHOP run.
        """
        return self._check_status(self.shop_run_event.external_id)

    def wait_until_complete(self) -> None:
        """
        Blocking call that waits until the until the SHOP run is complete.

        """
        while self.in_progress:
            logger.debug(f"{self.shop_run_event.external_id} is still running...")
            time.sleep(3)
        logger.debug(f"{self.shop_run_event.external_id} finished.")

    def get_results(self, wait: bool = True) -> ShopRunResult:
        """
        Retrieve the results of the SHOP run.

        Args:
            wait: If True, wait until the SHOP run is complete before retrieving the results.

        Returns:
            The results of the SHOP run.
        """
        if wait:
            if self.in_progress:
                logger.warning(f"{self.shop_run_event.external_id} is still running, waiting for results...")
            self.wait_until_complete()
        return self._retrieve_results(self)

    def __repr__(self) -> str:
        return f'<ShopRun status="{self.status}" event_external_id="{self.shop_run_event.external_id}">'

    def __str__(self) -> str:
        return self.__repr__()



from ._case import Case, CaseApply, CaseApplyList, CaseFields, CaseList, CaseTextFields
from ._commands_config import (
    CommandsConfig,
    CommandsConfigApply,
    CommandsConfigApplyList,
    CommandsConfigFields,
    CommandsConfigList,
    CommandsConfigTextFields,
)
from ._core import DomainModel, DomainModelApply
from ._file_ref import FileRef, FileRefApply, FileRefApplyList, FileRefFields, FileRefList, FileRefTextFields
from ._mapping import Mapping, MappingApply, MappingApplyList, MappingFields, MappingList, MappingTextFields
from ._model_template import (
    ModelTemplate,
    ModelTemplateApply,
    ModelTemplateApplyList,
    ModelTemplateFields,
    ModelTemplateList,
    ModelTemplateTextFields,
)
from ._processing_log import (
    ProcessingLog,
    ProcessingLogApply,
    ProcessingLogApplyList,
    ProcessingLogFields,
    ProcessingLogList,
    ProcessingLogTextFields,
)
from ._scenario import Scenario, ScenarioApply, ScenarioApplyList, ScenarioFields, ScenarioList, ScenarioTextFields
from ._transformation import (
    Transformation,
    TransformationApply,
    TransformationApplyList,
    TransformationFields,
    TransformationList,
    TransformationTextFields,
)

CaseApply.model_rebuild()
MappingApply.model_rebuild()
ModelTemplateApply.model_rebuild()
ScenarioApply.model_rebuild()

__all__ = [
    "DomainModel",
    "DomainModelApply",
    "Case",
    "CaseApply",
    "CaseList",
    "CaseApplyList",
    "CaseFields",
    "CaseTextFields",
    "CommandsConfig",
    "CommandsConfigApply",
    "CommandsConfigList",
    "CommandsConfigApplyList",
    "CommandsConfigFields",
    "CommandsConfigTextFields",
    "FileRef",
    "FileRefApply",
    "FileRefList",
    "FileRefApplyList",
    "FileRefFields",
    "FileRefTextFields",
    "Mapping",
    "MappingApply",
    "MappingList",
    "MappingApplyList",
    "MappingFields",
    "MappingTextFields",
    "ModelTemplate",
    "ModelTemplateApply",
    "ModelTemplateList",
    "ModelTemplateApplyList",
    "ModelTemplateFields",
    "ModelTemplateTextFields",
    "ProcessingLog",
    "ProcessingLogApply",
    "ProcessingLogList",
    "ProcessingLogApplyList",
    "ProcessingLogFields",
    "ProcessingLogTextFields",
    "Scenario",
    "ScenarioApply",
    "ScenarioList",
    "ScenarioApplyList",
    "ScenarioFields",
    "ScenarioTextFields",
    "Transformation",
    "TransformationApply",
    "TransformationList",
    "TransformationApplyList",
    "TransformationFields",
    "TransformationTextFields",
]



from __future__ import annotations

from collections.abc import Sequence
from typing import Generic, Literal, Optional, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.cogshop1.data_classes._core import T_TypeApplyNode, T_TypeNode, T_TypeNodeList

DEFAULT_LIMIT_READ = 25
INSTANCE_QUERY_LIMIT = 1_000
IN_FILTER_LIMIT = 5_000

Aggregations = Literal["avg", "count", "max", "min", "sum"]

_METRIC_AGGREGATIONS_BY_NAME = {
    "avg": dm.aggregations.Avg,
    "count": dm.aggregations.Count,
    "max": dm.aggregations.Max,
    "min": dm.aggregations.Min,
    "sum": dm.aggregations.Sum,
}


class TypeAPI(Generic[T_TypeNode, T_TypeApplyNode, T_TypeNodeList]):
    def __init__(
        self,
        client: CogniteClient,
        sources: dm.ViewIdentifier | Sequence[dm.ViewIdentifier] | dm.View | Sequence[dm.View],
        class_type: type[T_TypeNode],
        class_apply_type: type[T_TypeApplyNode],
        class_list: type[T_TypeNodeList],
    ):
        self._client = client
        self._sources = sources
        self._class_type = class_type
        self._class_apply_type = class_apply_type
        self._class_list = class_list

    @overload
    def _retrieve(self, external_id: str) -> T_TypeNode: ...

    @overload
    def _retrieve(self, external_id: Sequence[str]) -> T_TypeNodeList: ...

    def _retrieve(
        self, nodes: dm.NodeId | Sequence[dm.NodeId] | tuple[str, str] | Sequence[tuple[str, str]]
    ) -> T_TypeNode | T_TypeNodeList:
        is_multiple = (
            isinstance(nodes, Sequence)
            and not isinstance(nodes, str)
            and not (isinstance(nodes, tuple) and isinstance(nodes[0], str))
        )
        instances = self._client.data_modeling.instances.retrieve(nodes=nodes, sources=self._sources)
        if is_multiple:
            return self._class_list([self._class_type.from_node(node) for node in instances.nodes])
        return self._class_type.from_node(instances.nodes[0])

    def _search(
        self,
        view_id: dm.ViewId,
        query: str,
        properties_by_field: dict[str, str],
        properties: str | Sequence[str],
        filter_: dm.Filter | None = None,
        limit: int = DEFAULT_LIMIT_READ,
    ) -> T_TypeNodeList:
        if isinstance(properties, str):
            properties = [properties]

        if properties:
            properties = [properties_by_field.get(prop, prop) for prop in properties]

        nodes = self._client.data_modeling.instances.search(view_id, query, "node", properties, filter_, limit)
        return self._class_list([self._class_type.from_node(node) for node in nodes])

    @overload
    def _aggregate(
        self,
        view_id: dm.ViewId,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        properties_by_field: dict[str, str],
        properties: str | Sequence[str] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: str | Sequence[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def _aggregate(
        self,
        view_id: dm.ViewId,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        properties_by_field: dict[str, str],
        properties: Optional[str | Sequence[str]] = None,
        group_by: str | Sequence[str] | None = None,
        query: str | None = None,
        search_properties: str | Sequence[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def _aggregate(
        self,
        view_id: dm.ViewId,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        properties_by_field: dict[str, str],
        properties: str | Sequence[str] | None = None,
        group_by: str | Sequence[str] | None = None,
        query: str | None = None,
        search_properties: str | Sequence[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        if isinstance(group_by, str):
            group_by = [group_by]

        if group_by:
            group_by = [properties_by_field.get(prop, prop) for prop in group_by]

        if isinstance(search_properties, str):
            search_properties = [search_properties]

        if search_properties:
            search_properties = [properties_by_field.get(prop, prop) for prop in search_properties]

        if isinstance(properties, str):
            properties = [properties]

        if properties:
            properties = [properties_by_field.get(prop, prop) for prop in properties]

        if isinstance(aggregate, (str, dm.aggregations.MetricAggregation)):
            aggregate = [aggregate]

        if properties is None and (invalid := [agg for agg in aggregate if isinstance(agg, str) and agg != "count"]):
            raise ValueError(f"Cannot aggregate on {invalid} without specifying properties")

        aggregates = []
        for agg in aggregate:
            if isinstance(agg, dm.aggregations.MetricAggregation):
                aggregates.append(agg)
            elif isinstance(agg, str):
                if agg == "count" and properties is None:
                    # Special case for count, we just pick the first property
                    first_prop = next(iter(properties_by_field.values()))
                    aggregates.append(dm.aggregations.Count(first_prop))
                elif properties is None:
                    raise ValueError(f"Cannot aggregate on {agg} without specifying properties")
                else:
                    for prop in properties:
                        aggregates.append(_METRIC_AGGREGATIONS_BY_NAME[agg](prop))
            else:
                raise TypeError(f"Expected str or MetricAggregation, got {type(agg)}")

        result = self._client.data_modeling.instances.aggregate(
            view_id, aggregates, "node", group_by, query, search_properties, filter, limit
        )
        if group_by is None:
            return result[0].aggregates
        return result

    def _histogram(
        self,
        view_id: dm.ViewId,
        property: str,
        interval: float,
        properties_by_field: dict[str, str],
        query: str | None = None,
        search_properties: str | Sequence[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        property = properties_by_field.get(property, property)

        if isinstance(search_properties, str):
            search_properties = [search_properties]
        if search_properties:
            search_properties = [properties_by_field.get(prop, prop) for prop in search_properties]

        return self._client.data_modeling.instances.histogram(
            view_id, dm.aggregations.Histogram(property, interval), "node", query, search_properties, filter, limit
        )

    def _list(self, limit: int = DEFAULT_LIMIT_READ, filter: dm.Filter | None = None) -> T_TypeNodeList:
        nodes = self._client.data_modeling.instances.list("node", sources=self._sources, limit=limit, filter=filter)
        return self._class_list([self._class_type.from_node(node) for node in nodes])



from __future__ import annotations

import datetime
import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._alert import Alert, AlertGraphQL, AlertWrite
    from ._bid_matrix import BidMatrix, BidMatrixGraphQL, BidMatrixWrite
    from ._bid_method import BidMethod, BidMethodGraphQL, BidMethodWrite
    from ._price_area import PriceArea, PriceAreaGraphQL, PriceAreaWrite


__all__ = [
    "BidDocument",
    "BidDocumentWrite",
    "BidDocumentApply",
    "BidDocumentList",
    "BidDocumentWriteList",
    "BidDocumentApplyList",
    "BidDocumentFields",
    "BidDocumentTextFields",
]


BidDocumentTextFields = Literal["name"]
BidDocumentFields = Literal["name", "delivery_date", "start_calculation", "end_calculation", "is_complete"]

_BIDDOCUMENT_PROPERTIES_BY_FIELD = {
    "name": "name",
    "delivery_date": "deliveryDate",
    "start_calculation": "startCalculation",
    "end_calculation": "endCalculation",
    "is_complete": "isComplete",
}


class BidDocumentGraphQL(GraphQLCore):
    """This represents the reading version of bid document, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid document.
        data_record: The data record of the bid document node.
        name: Unique name for a given instance of a Bid Document. A combination of name, priceArea, date and startCalculation.
        delivery_date: The date of the Bid.
        start_calculation: Timestamp of when the Bid calculation workflow started.
        end_calculation: Timestamp of when the Bid calculation workflow completed.
        is_complete: Indicates that the Bid calculation workflow has completed (although has not necessarily succeeded).
        alerts: An array of calculation level Alerts.
        price_area: The price area field.
        method: The method field.
        total: The total field.
        partials: The partial field.
    """

    view_id = dm.ViewId("power-ops-day-ahead-bid", "BidDocument", "1")
    name: Optional[str] = None
    delivery_date: Optional[datetime.date] = Field(None, alias="deliveryDate")
    start_calculation: Optional[datetime.datetime] = Field(None, alias="startCalculation")
    end_calculation: Optional[datetime.datetime] = Field(None, alias="endCalculation")
    is_complete: Optional[bool] = Field(None, alias="isComplete")
    alerts: Optional[list[AlertGraphQL]] = Field(default=None, repr=False)
    price_area: Optional[PriceAreaGraphQL] = Field(None, repr=False, alias="priceArea")
    method: Optional[BidMethodGraphQL] = Field(None, repr=False)
    total: Optional[BidMatrixGraphQL] = Field(None, repr=False)
    partials: Optional[list[BidMatrixGraphQL]] = Field(default=None, repr=False)

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("alerts", "price_area", "method", "total", "partials", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> BidDocument:
        """Convert this GraphQL format of bid document to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return BidDocument(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            delivery_date=self.delivery_date,
            start_calculation=self.start_calculation,
            end_calculation=self.end_calculation,
            is_complete=self.is_complete,
            alerts=[alert.as_read() if isinstance(alert, GraphQLCore) else alert for alert in self.alerts or []],
            price_area=self.price_area.as_read() if isinstance(self.price_area, GraphQLCore) else self.price_area,
            method=self.method.as_read() if isinstance(self.method, GraphQLCore) else self.method,
            total=self.total.as_read() if isinstance(self.total, GraphQLCore) else self.total,
            partials=[
                partial.as_read() if isinstance(partial, GraphQLCore) else partial for partial in self.partials or []
            ],
        )

    def as_write(self) -> BidDocumentWrite:
        """Convert this GraphQL format of bid document to the writing format."""
        return BidDocumentWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            delivery_date=self.delivery_date,
            start_calculation=self.start_calculation,
            end_calculation=self.end_calculation,
            is_complete=self.is_complete,
            alerts=[alert.as_write() if isinstance(alert, DomainModel) else alert for alert in self.alerts or []],
            price_area=self.price_area.as_write() if isinstance(self.price_area, DomainModel) else self.price_area,
            method=self.method.as_write() if isinstance(self.method, DomainModel) else self.method,
            total=self.total.as_write() if isinstance(self.total, DomainModel) else self.total,
            partials=[
                partial.as_write() if isinstance(partial, DomainModel) else partial for partial in self.partials or []
            ],
        )


class BidDocument(DomainModel):
    """This represents the reading version of bid document.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid document.
        data_record: The data record of the bid document node.
        name: Unique name for a given instance of a Bid Document. A combination of name, priceArea, date and startCalculation.
        delivery_date: The date of the Bid.
        start_calculation: Timestamp of when the Bid calculation workflow started.
        end_calculation: Timestamp of when the Bid calculation workflow completed.
        is_complete: Indicates that the Bid calculation workflow has completed (although has not necessarily succeeded).
        alerts: An array of calculation level Alerts.
        price_area: The price area field.
        method: The method field.
        total: The total field.
        partials: The partial field.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power-ops-types", "DayAheadBidDocument"
    )
    name: Optional[str] = None
    delivery_date: datetime.date = Field(alias="deliveryDate")
    start_calculation: Optional[datetime.datetime] = Field(None, alias="startCalculation")
    end_calculation: Optional[datetime.datetime] = Field(None, alias="endCalculation")
    is_complete: Optional[bool] = Field(None, alias="isComplete")
    alerts: Union[list[Alert], list[str], list[dm.NodeId], None] = Field(default=None, repr=False)
    price_area: Union[PriceArea, str, dm.NodeId, None] = Field(None, repr=False, alias="priceArea")
    method: Union[BidMethod, str, dm.NodeId, None] = Field(None, repr=False)
    total: Union[BidMatrix, str, dm.NodeId, None] = Field(None, repr=False)
    partials: Union[list[BidMatrix], list[str], list[dm.NodeId], None] = Field(default=None, repr=False)

    def as_write(self) -> BidDocumentWrite:
        """Convert this read version of bid document to the writing version."""
        return BidDocumentWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            delivery_date=self.delivery_date,
            start_calculation=self.start_calculation,
            end_calculation=self.end_calculation,
            is_complete=self.is_complete,
            alerts=[alert.as_write() if isinstance(alert, DomainModel) else alert for alert in self.alerts or []],
            price_area=self.price_area.as_write() if isinstance(self.price_area, DomainModel) else self.price_area,
            method=self.method.as_write() if isinstance(self.method, DomainModel) else self.method,
            total=self.total.as_write() if isinstance(self.total, DomainModel) else self.total,
            partials=[
                partial.as_write() if isinstance(partial, DomainModel) else partial for partial in self.partials or []
            ],
        )

    def as_apply(self) -> BidDocumentWrite:
        """Convert this read version of bid document to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidDocumentWrite(DomainModelWrite):
    """This represents the writing version of bid document.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid document.
        data_record: The data record of the bid document node.
        name: Unique name for a given instance of a Bid Document. A combination of name, priceArea, date and startCalculation.
        delivery_date: The date of the Bid.
        start_calculation: Timestamp of when the Bid calculation workflow started.
        end_calculation: Timestamp of when the Bid calculation workflow completed.
        is_complete: Indicates that the Bid calculation workflow has completed (although has not necessarily succeeded).
        alerts: An array of calculation level Alerts.
        price_area: The price area field.
        method: The method field.
        total: The total field.
        partials: The partial field.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power-ops-types", "DayAheadBidDocument"
    )
    name: Optional[str] = None
    delivery_date: datetime.date = Field(alias="deliveryDate")
    start_calculation: Optional[datetime.datetime] = Field(None, alias="startCalculation")
    end_calculation: Optional[datetime.datetime] = Field(None, alias="endCalculation")
    is_complete: Optional[bool] = Field(None, alias="isComplete")
    alerts: Union[list[AlertWrite], list[str], list[dm.NodeId], None] = Field(default=None, repr=False)
    price_area: Union[PriceAreaWrite, str, dm.NodeId, None] = Field(None, repr=False, alias="priceArea")
    method: Union[BidMethodWrite, str, dm.NodeId, None] = Field(None, repr=False)
    total: Union[BidMatrixWrite, str, dm.NodeId, None] = Field(None, repr=False)
    partials: Union[list[BidMatrixWrite], list[str], list[dm.NodeId], None] = Field(default=None, repr=False)

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            BidDocument, dm.ViewId("power-ops-day-ahead-bid", "BidDocument", "1")
        )

        properties: dict[str, Any] = {}

        if self.name is not None or write_none:
            properties["name"] = self.name

        if self.delivery_date is not None:
            properties["deliveryDate"] = self.delivery_date.isoformat() if self.delivery_date else None

        if self.start_calculation is not None or write_none:
            properties["startCalculation"] = (
                self.start_calculation.isoformat(timespec="milliseconds") if self.start_calculation else None
            )

        if self.end_calculation is not None or write_none:
            properties["endCalculation"] = (
                self.end_calculation.isoformat(timespec="milliseconds") if self.end_calculation else None
            )

        if self.is_complete is not None or write_none:
            properties["isComplete"] = self.is_complete

        if self.price_area is not None:
            properties["priceArea"] = {
                "space": self.space if isinstance(self.price_area, str) else self.price_area.space,
                "externalId": self.price_area if isinstance(self.price_area, str) else self.price_area.external_id,
            }

        if self.method is not None:
            properties["method"] = {
                "space": self.space if isinstance(self.method, str) else self.method.space,
                "externalId": self.method if isinstance(self.method, str) else self.method.external_id,
            }

        if self.total is not None:
            properties["total"] = {
                "space": self.space if isinstance(self.total, str) else self.total.space,
                "externalId": self.total if isinstance(self.total, str) else self.total.external_id,
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power-ops-types", "calculationIssue")
        for alert in self.alerts or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=alert,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        edge_type = dm.DirectRelationReference("power-ops-types", "partialBid")
        for partial in self.partials or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=partial,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.price_area, DomainModelWrite):
            other_resources = self.price_area._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.method, DomainModelWrite):
            other_resources = self.method._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.total, DomainModelWrite):
            other_resources = self.total._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class BidDocumentApply(BidDocumentWrite):
    def __new__(cls, *args, **kwargs) -> BidDocumentApply:
        warnings.warn(
            "BidDocumentApply is deprecated and will be removed in v1.0. Use BidDocumentWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "BidDocument.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class BidDocumentList(DomainModelList[BidDocument]):
    """List of bid documents in the read version."""

    _INSTANCE = BidDocument

    def as_write(self) -> BidDocumentWriteList:
        """Convert these read versions of bid document to the writing versions."""
        return BidDocumentWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> BidDocumentWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidDocumentWriteList(DomainModelWriteList[BidDocumentWrite]):
    """List of bid documents in the writing version."""

    _INSTANCE = BidDocumentWrite


class BidDocumentApplyList(BidDocumentWriteList): ...


def _create_bid_document_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    min_delivery_date: datetime.date | None = None,
    max_delivery_date: datetime.date | None = None,
    min_start_calculation: datetime.datetime | None = None,
    max_start_calculation: datetime.datetime | None = None,
    min_end_calculation: datetime.datetime | None = None,
    max_end_calculation: datetime.datetime | None = None,
    is_complete: bool | None = None,
    price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    total: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if min_delivery_date is not None or max_delivery_date is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("deliveryDate"),
                gte=min_delivery_date.isoformat() if min_delivery_date else None,
                lte=max_delivery_date.isoformat() if max_delivery_date else None,
            )
        )
    if min_start_calculation is not None or max_start_calculation is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("startCalculation"),
                gte=min_start_calculation.isoformat(timespec="milliseconds") if min_start_calculation else None,
                lte=max_start_calculation.isoformat(timespec="milliseconds") if max_start_calculation else None,
            )
        )
    if min_end_calculation is not None or max_end_calculation is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("endCalculation"),
                gte=min_end_calculation.isoformat(timespec="milliseconds") if min_end_calculation else None,
                lte=max_end_calculation.isoformat(timespec="milliseconds") if max_end_calculation else None,
            )
        )
    if isinstance(is_complete, bool):
        filters.append(dm.filters.Equals(view_id.as_property_ref("isComplete"), value=is_complete))
    if price_area and isinstance(price_area, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("priceArea"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": price_area}
            )
        )
    if price_area and isinstance(price_area, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("priceArea"), value={"space": price_area[0], "externalId": price_area[1]}
            )
        )
    if price_area and isinstance(price_area, list) and isinstance(price_area[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("priceArea"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in price_area],
            )
        )
    if price_area and isinstance(price_area, list) and isinstance(price_area[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("priceArea"),
                values=[{"space": item[0], "externalId": item[1]} for item in price_area],
            )
        )
    if method and isinstance(method, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("method"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": method}
            )
        )
    if method and isinstance(method, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("method"), value={"space": method[0], "externalId": method[1]})
        )
    if method and isinstance(method, list) and isinstance(method[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("method"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in method],
            )
        )
    if method and isinstance(method, list) and isinstance(method[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("method"), values=[{"space": item[0], "externalId": item[1]} for item in method]
            )
        )
    if total and isinstance(total, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("total"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": total}
            )
        )
    if total and isinstance(total, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("total"), value={"space": total[0], "externalId": total[1]})
        )
    if total and isinstance(total, list) and isinstance(total[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("total"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in total],
            )
        )
    if total and isinstance(total, list) and isinstance(total[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("total"), values=[{"space": item[0], "externalId": item[1]} for item in total]
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import json
from typing import Any

import pandas as pd
from cognite.client.data_classes import Sequence

from cognite.powerops.resync import config
from cognite.powerops.resync.models._shared_v1_v2.market_model import _map_price_scenarios_by_name
from cognite.powerops.resync.models.base import CDFSequence
from cognite.powerops.resync.models.v1.market import (
    BenchmarkBid,
    BenchmarkProcess,
    DayAheadBid,
    DayAheadProcess,
    MarketModel,
    NordPoolMarket,
    RKOMBid,
    RKOMBidCombination,
    RKOMCombinationBid,
    RKOMMarket,
    RKOMPlants,
    RKOMProcess,
    ShopTransformation,
)
from cognite.powerops.resync.models.v1.production import PriceArea


def to_market_asset_model(
    configuration: config.MarketConfig, price_areas: list[PriceArea], market_name: str
) -> MarketModel:
    nord_pool = NordPoolMarket(**configuration.market.model_dump())
    # The external_id is not following the naming convention, so we need to set it manually
    nord_pool.external_id = configuration.market.external_id

    benchmarking_processes, benchmarking_deferred_attrs = _to_benchmarking_process(configuration.benchmarks)

    dayahead_processes = _to_dayahead_process(
        bid_process_configs=configuration.bidprocess,
        bidmatrix_generators=configuration.bidmatrix_generators,
        price_scenarios_by_id=configuration.price_scenario_by_id,
        benchmarking=benchmarking_processes[0],
        price_areas=price_areas,
    )

    benchmarking_processes = _resolve_deferred_attrs(
        benchmarking_processes=benchmarking_processes,
        deferred_attrs=benchmarking_deferred_attrs,
        dayahead_processes=dayahead_processes,
    )

    rkom = _to_rkom_market(
        rkom_bid_process=configuration.rkom_bid_process,
        price_scenarios_by_id=configuration.price_scenario_by_id,
        market_name=market_name,
        rkom_market_config=configuration.rkom_market,
        rkom_bid_combination_configs=configuration.rkom_bid_combination,
    )

    model = MarketModel(
        nordpool_market=[nord_pool],
        rkom_market=rkom.rkom_market,
        benchmark_processes=benchmarking_processes,
        dayahead_processes=dayahead_processes,
        rkom_processes=rkom.rkom_processes,
        combinations=rkom.combinations,
    )

    return model


_DeferredAttrsT = dict[str, dict[str, Any]]


def _to_benchmarking_process(
    benchmarks: list[config.BenchmarkingConfig],
) -> tuple[list[BenchmarkProcess], _DeferredAttrsT]:
    benchmarkings: list[BenchmarkProcess] = []
    if len(benchmarks) > 1:
        # The external id for the benchmarking is hardcoded and thus there can be only one.
        raise NotImplementedError("Only one benchmarking config is supported")
    deferred_attrs = {}
    for benchmarking_config in benchmarks:
        bid = BenchmarkBid(
            date=json.dumps(benchmarking_config.bid_date.operations),
            market_config_external_id=benchmarking_config.market_config_external_id,
        )
        process = BenchmarkProcess(
            name="Benchmarking config DA",
            bid=bid,
            shop=ShopTransformation(
                starttime=json.dumps(benchmarking_config.shop_start.operations),
                endtime=json.dumps(benchmarking_config.shop_end.operations),
            ),
            production_plan_time_series=(
                json.dumps(benchmarking_config.production_plan_time_series, ensure_ascii=False)
                if benchmarking_config.production_plan_time_series
                else []
            ),  # ensure_ascii=False to treat Nordic letters properly,
            benchmarking_metrics=benchmarking_config.relevant_shop_objective_metrics,
            bid_process_configuration_assets=[],
        )
        # The external_id is not following the naming convention, so we need to set it manually
        process.external_id = "POWEROPS_dayahead_bidding_benchmarking_config"
        benchmarkings.append(process)
        deferred_attrs[process.external_id] = {
            "bid_process_configuration_assets": [
                f"POWEROPS_bid_process_configuration_{name}"
                for name in benchmarking_config.bid_process_configuration_assets
            ],
        }

    return benchmarkings, deferred_attrs


def _resolve_deferred_attrs(
    benchmarking_processes: list[BenchmarkProcess],
    deferred_attrs: _DeferredAttrsT,
    dayahead_processes: list[DayAheadProcess],
) -> list[BenchmarkProcess]:
    for external_id, attrs in deferred_attrs.items():
        benchmarking_process = next(p for p in benchmarking_processes if p.external_id == external_id)
        for attr, deferred_value in attrs.items():
            if attr == "bid_process_configuration_assets":
                value = [p for ext_id in deferred_value for p in dayahead_processes if p.external_id == ext_id]
            else:
                raise NotImplementedError()
            setattr(benchmarking_process, attr, value)
    return benchmarking_processes


def _to_dayahead_process(
    bid_process_configs: list[config.BidProcessConfig],
    bidmatrix_generators: list[config.BidMatrixGeneratorConfig],
    price_scenarios_by_id: dict[str, config.PriceScenario],
    benchmarking: BenchmarkProcess,
    price_areas: list[PriceArea],
) -> list[DayAheadProcess]:
    dayahead_processes: list[DayAheadProcess] = []
    bidmatrix_generators_by_name: dict[str, config.BidMatrixGeneratorConfig] = {g.name: g for g in bidmatrix_generators}
    for process in bid_process_configs:
        price_scenarios_by_name = _map_price_scenarios_by_name(
            process.price_scenarios, price_scenarios_by_id, config.MARKET_BY_PRICE_AREA[process.price_area_name]
        )

        price_area = next((pa for pa in price_areas if pa.name == process.price_area_name), None)
        if not price_area:
            raise ValueError(f"Price area {process.price_area_name} not found")

        # Create bid matrix generator config sequence
        bidmatrix_generator = bidmatrix_generators_by_name[process.bid_matrix_generator]
        column_def = [
            {"valueType": "STRING", "externalId": external_id}
            for external_id in bidmatrix_generator.column_external_ids
        ]
        bid_matrix_generator_sequence = Sequence(
            external_id=f"POWEROPS_bid_matrix_generator_config_{process.name}",
            name=f"POWEROPS bid matrix generator config {process.name}",
            description="Configuration of bid matrix generation method to use for each plant in the price area",
            columns=column_def,
            metadata={
                # TODO: Rename this from "shop:type" to something without "shop"
                #   (but check if e.g. power-ops-functions uses this)
                "bid:price_area": f"price_area_{process.price_area_name}",
                "shop:type": "bid_matrix_generator_config",
                "bid:bid_process_configuration_name": process.name,
            },
        )
        bid_matrix_generator_sequence_content = pd.DataFrame(
            [
                [plant.name, bidmatrix_generator.default_method, bidmatrix_generator.default_function_external_id]
                for plant in price_area.plants
            ],
            columns=bidmatrix_generator.column_external_ids,
        )

        # Create incremental mapping sequences
        incremental_mapping_sequences = []
        for watercourse in price_area.watercourses:
            price_scenarios: dict[str, config.PriceScenario] = price_scenarios_by_name
            if process.price_scenarios_per_watercourse:
                try:
                    price_scenarios = {
                        scenario_name: price_scenarios[scenario_name]
                        for scenario_name in process.price_scenarios_per_watercourse[watercourse.name]
                    }
                except KeyError as e:
                    raise KeyError(
                        f"Watercourse {watercourse.name} not defined in price_scenarios_per_watercourse "
                        f"for BidProcessConfig {process.name}"
                    ) from e
            for scenario_name, price_scenario in price_scenarios.items():
                time_series_mapping = price_scenario.to_time_series_mapping()
                metadata = {
                    "shop:watercourse": watercourse.name,
                    "shop:type": "incremental_mapping",
                    "bid:scenario_name": scenario_name,
                }
                external_id = f"SHOP_{watercourse.name}_incremental_mapping_{process.name}_{scenario_name}"
                name = external_id.replace("_", " ")
                sequence = Sequence(
                    name=name,
                    external_id=external_id,
                    description="Mapping between SHOP paths and CDF TimeSeries",
                    columns=time_series_mapping.column_definitions,
                    metadata=metadata,
                )
                incremental_mapping = CDFSequence(sequence=sequence, content=time_series_mapping.to_dataframe())
                incremental_mapping_sequences.append(incremental_mapping)

        # Create the DayAheadBidProcess Data Class
        dayahead_process = DayAheadProcess(
            name=f"Bid process configuration {process.name}",
            description="Bid process configuration defining how to run a bid matrix generation process",
            bid=DayAheadBid(
                date=str(process.bid_date or benchmarking.bid.date),
                market_config_external_id=benchmarking.bid.market_config_external_id,
                is_default_config_for_price_area=process.is_default_config_for_price_area,
                main_scenario=process.main_scenario,
                price_area=f"price_area_{process.price_area_name}",
                price_scenarios={
                    scenario_name: scenario.time_series_external_id
                    for scenario_name, scenario in price_scenarios_by_name.items()
                },
                no_shop=process.no_shop,
                bid_process_configuration_name=process.name,
                bid_matrix_generator_config_external_id=f"POWEROPS_bid_matrix_generator_config_{process.name}",
            ),
            shop=ShopTransformation(
                starttime=str(process.shop_start or benchmarking.shop.starttime),
                endtime=str(process.shop_end or benchmarking.shop.endtime),
            ),
            bid_matrix_generator_config=CDFSequence(
                sequence=bid_matrix_generator_sequence, content=bid_matrix_generator_sequence_content
            ),
            incremental_mapping=incremental_mapping_sequences,
        )
        # The IDs are inconsistently compared to the other data classes, so we need to set them manually
        dayahead_process.external_id = f"POWEROPS_bid_process_configuration_{process.name}"
        dayahead_process.type_ = "POWEROPS_bid_process_configuration"

        dayahead_processes.append(dayahead_process)

    return dayahead_processes


def _to_rkom_market(
    rkom_bid_process: list[config.RKOMBidProcessConfig],
    price_scenarios_by_id: dict[str, config.PriceScenario],
    market_name: str,
    rkom_market_config: config.RkomMarketConfig,
    rkom_bid_combination_configs: list[config.RKOMBidCombinationConfig] | None = None,
) -> MarketModel:
    model = MarketModel()

    rkom_market = RKOMMarket(**rkom_market_config.model_dump())
    rkom_market.external_id = rkom_market_config.external_id
    model.rkom_market.append(rkom_market)

    for configuration in rkom_bid_process:
        price_scenarios_by_name = _map_price_scenarios_by_name(
            configuration.price_scenarios, price_scenarios_by_id, market_name
        )
        incremental_mapping_sequences = []
        for scenario_name, price_scenario in price_scenarios_by_name.items():
            mappings = price_scenario.to_time_series_mapping()
            for reserve_scenario in configuration.reserve_scenarios:
                reserve_mapping = reserve_scenario.to_time_series_mapping()
                metadata = {
                    "shop:watercourse": configuration.watercourse,
                    "shop:type": "rkom_incremental_mapping",
                    "bid:scenario_name": scenario_name,
                    "bid:reserve_volume": str(reserve_scenario.volume),
                }
                external_id = (
                    f"SHOP_{configuration.watercourse}_incremental_mapping_"
                    f"{configuration.name}_{scenario_name}_{reserve_scenario.volume}MW"
                )
                name = f"{configuration.watercourse} {scenario_name} {reserve_scenario.volume} MW"
                sequence = Sequence(
                    name=name,
                    external_id=external_id,
                    description="Mapping between SHOP paths and CDF TimeSeries",
                    columns=(mappings + reserve_mapping).column_definitions,
                    metadata=metadata,
                )
                content = (mappings + reserve_mapping).to_dataframe()
                incremental_mapping = CDFSequence(sequence=sequence, content=content)
                incremental_mapping_sequences.append(incremental_mapping)

        process = RKOMProcess(
            name=configuration.name,
            description=f"RKOM bid generation config for {configuration.watercourse}",
            bid=RKOMBid(
                date=str(configuration.bid_date),
                auction=configuration.reserve_scenarios.auction.value,
                block=configuration.reserve_scenarios.block,
                product=configuration.reserve_scenarios.product,
                watercourse=configuration.watercourse,
                method=configuration.method,
                minimum_price=str(configuration.minimum_price),
                price_premium=str(configuration.price_premium),
                price_scenarios=json.dumps(list(price_scenarios_by_name)),
                reserve_scenarios=str(configuration.reserve_scenarios),
            ),
            shop=ShopTransformation(starttime=str(configuration.shop_start), endtime=str(configuration.shop_end)),
            timezone=configuration.timezone,
            rkom=RKOMPlants(plants=json.dumps(sorted(configuration.rkom_plants))),
            incremental_mapping=incremental_mapping_sequences,
        )
        process.external_id = configuration.external_id
        process.type_ = "POWEROPS"
        model.rkom_processes.append(process)

    for configuration in rkom_bid_combination_configs or []:
        sequence_external_id = f"RKOM_bid_combination_configuration_{configuration.auction.value}_{configuration.name}"
        combination = RKOMBidCombination(
            name=sequence_external_id.replace("_", " "),
            description="Defining which RKOM bid methods should be combined (into the total bid form)",
            bid=RKOMCombinationBid(
                auction=configuration.auction.value,
                combination_name=configuration.name,
                rkom_bid_configs=configuration.rkom_bid_config_external_ids,
            ),
        )
        combination.external_id = sequence_external_id
        model.combinations.append(combination)

    return model



"""
This module contains the main functions for the resync tool.
"""

from __future__ import annotations

import itertools
import logging
import os
import re
import tempfile
from pathlib import Path
from typing import Any, Optional, cast

from click.core import Command
from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import DataModelId, MappedProperty, ViewList
from cognite.client.exceptions import CogniteAPIError
from cognite_toolkit._cdf_tk.load import DataModelLoader as ToolkitDataModelLoader  # type: ignore[import-untyped]
from cognite_toolkit._cdf_tk.load import ViewLoader  # type: ignore[import-untyped]  # type: ignore[import-untyped]
from cognite_toolkit._cdf_tk.utils import CDFToolConfig, calculate_directory_hash  # type: ignore[import-untyped]
from rich import print
from rich.panel import Panel
from typer import Context
from yaml import safe_dump

from cognite.powerops.cdf_labels import AssetLabel, RelationshipLabel
from cognite.powerops.client.powerops_client import PowerOpsClient
from cognite.powerops.resync import diff, models
from cognite.powerops.resync.config import ReSyncConfig
from cognite.powerops.resync.diff import FieldDifference, ModelDifference, ModelDifferences
from cognite.powerops.resync.models.base import AssetModel, CDFFile, CDFSequence, DataModel, Model, SpaceId
from cognite.powerops.resync.models.v2.powerops_models import DataModelLoader
from cognite.powerops.utils.serialization import environment_variables

from .cdf import get_cognite_api
from .transform import transform
from .validation import _clean_relationships

os.environ["SENTRY_ENABLED"] = "false"
# ruff: noqa: E402
from cognite_toolkit._cdf import Common, build, deploy  # type: ignore[import-untyped]

logger = logging.getLogger(__name__)

MODELS_BY_NAME = {m.__name__: m for m in itertools.chain(models.V1_MODELS, models.V2_MODELS)}

DATAMODEL_ID_TO_RESYNC_NAME: dict[DataModelId, str] = {
    m.data_model_ids()[0]: m.__name__
    for m in itertools.chain(models.V1_MODELS, models.V2_MODELS)
    if issubclass(m, DataModel) and m.data_model_ids()
}


def init(client: PowerOpsClient | None, is_dev: bool = False, dry_run: bool = False, verbose: bool = False) -> None:
    """
    This function will create the data models in CDF that are required for resync to work. It will not overwrite
    existing models.

    Args:
        client: The PowerOpsClient to use. If not provided, a new client will be created.
        is_dev: Whether the deployment is for development environment. If true, the models views and data models
                will be deleted and recreated.
        dry_run: Whether to run the command as a dry run, meaning no resources will be created.
    """
    client = client or PowerOpsClient.from_settings()
    cdf = client.cdf
    result = re.findall(r"^https://([\w-]+).cognitedata.com", cdf.config.base_url)
    if result:
        cluster = result[0]
    else:
        raise ValueError(f"Invalid base_url {cdf.config.base_url}")

    with environment_variables(
        {
            "CDF_CLUSTER": cluster,
            "CDF_URL": cdf.config.base_url,
            "IDP_CLIENT_ID": cdf.config.credentials.client_id,  # type: ignore[attr-defined]
            "IDP_CLIENT_SECRET": "dummy",
            "IDP_TOKEN_URL": (
                cdf.config.credentials.authority_url  # type: ignore[attr-defined]
                if hasattr(cdf.config.credentials, "authority_url")
                else cdf.config.credentials.token_url  # type: ignore[attr-defined]
            ),
            "CDF_PROJECT": cdf.config.project,
        }
    ):
        tool_config = CDFToolConfig()
        tool_config._client = cdf

    powerops_folder = Path(__file__).parent.parent.parent
    with environment_variables({"SENTRY_ENABLED": "false"}):
        ctx = Context(Command("build"))
        ctx.obj = Common(
            override_env=False,
            verbose=verbose,
            cluster=cluster,
            project=cdf.config.project,
            mockToolGlobals=tool_config,
        )
        build(ctx=ctx, source_dir=str(powerops_folder), build_dir="build", build_env="dev", clean=True)

    build_folder = Path.cwd() / "build"

    loader = DataModelLoader(build_folder)
    schema = loader.load()
    logger.info("Loaded all powerops data models")
    # TODO: fix validation
    DataModelLoader.validate(schema)
    logger.info("Validated all powerops data models")

    # Nodes should not be deployed, so we remove them from the build folder
    for node_file in build_folder.glob("data_models/*.powerops_nodes.yaml"):
        node_file.unlink()

    build_hash = calculate_directory_hash(build_folder)
    filepath = Path(tempfile.gettempdir()) / "powerops_init_command.txt"
    if dry_run:
        filepath.write_text(build_hash)
    elif filepath.exists() and filepath.read_text() == build_hash:
        ...
    else:
        print(
            Panel(
                "[bold red]Error: [/] `powerops init` has not been run with --dry-run before running without it."
                "Please run `powerops init` with --dry-run first to verify the changes.",
                title="No dry-run",
            )
        )
        exit(1)

    if is_dev:
        changed = loader.changed_views(cdf, schema, verbose)
        if changed:
            print(Panel(f"Detected {len(changed)} changed views"))
            if verbose:
                for view in changed:
                    logger.info(f"Changed view: {view}")
            dependencies = {view_id for dependencies in changed.values() for view_id in dependencies}
            print(f"Detected {len(dependencies)} dependent views")
            prefix = "Would delete" if dry_run else "Deleting"
            to_delete = list(dependencies | set(changed))
            print(f"{prefix} changed and dependent {len(to_delete)} views")

            if not dry_run:
                deleted = cdf.data_modeling.views.delete(to_delete)
                print(f"Deleted {len(deleted)} views")

            data_model_ids_to_delete = set(loader.dependent_data_models(schema, set(to_delete)))
        else:
            data_model_ids_to_delete = set()
            print(Panel("No changes detected in any views"))

        existing = cdf.data_modeling.data_models.retrieve(schema.data_models.as_ids())
        existing_by_id = {model.as_id(): model for model in existing}
        loader = ToolkitDataModelLoader.create_loader(tool_config)
        for local_data_model in schema.data_models:
            if local_data_model.as_id() not in existing_by_id:
                continue
            existing_data_model = existing_by_id[local_data_model.as_id()]

            if not loader._is_equal_custom(local_data_model, existing_data_model):  # type: ignore[attr-defined]
                data_model_ids_to_delete.add(local_data_model.as_id())

        print(Panel(f"Detected {len(data_model_ids_to_delete)} dependent and changed data models"))
        prefix = "Would delete" if dry_run else "Deleting"
        print(f"{prefix} dependent {len(data_model_ids_to_delete)} data models")
        if not dry_run:
            deleted_models = cdf.data_modeling.data_models.delete(list(data_model_ids_to_delete))
            print(f"Deleted {len(deleted_models)} data models")

    with environment_variables({"SENTRY_ENABLED": "false"}):
        ctx = Context(Command("deploy"))
        ctx.obj = Common(
            override_env=False,
            verbose=verbose,
            cluster=cluster,
            project=cdf.config.project,
            mockToolGlobals=tool_config,
        )
        try:
            deploy(
                ctx,
                build_dir="build",
                build_env="dev",
                interactive=False,
                drop=False,
                drop_data=False,
                dry_run=dry_run,
                include=None,
            )
        except SystemExit as e:
            if e.code != 0 and not dry_run:
                # Toolkit currently calls apply once for each view, which typically fails due to dependencies.
                # We try to deploy the views first, and then retry the deploy.
                print(Panel("Trying deploying views first", title="Deploy failed"))
                view_loader = ViewLoader.create_loader(tool_config)
                view_files = view_loader.find_files(build_folder / "data_models")
                views = [
                    view_loader.load_resource(view_file, tool_config, skip_validation=False) for view_file in view_files
                ]
                created = cdf.data_modeling.views.apply(views)
                print(f"Created {len(created)} views. Retrying deploy...")
                deploy(
                    ctx,
                    build_dir="build",
                    build_env="dev",
                    interactive=False,
                    drop=False,
                    drop_data=False,
                    dry_run=dry_run,
                    include=None,
                )
            else:
                raise


def validate(config_dir: str | Path, market: str) -> Any:
    raise NotImplementedError("validate is not implemented")


# def validate(config_dir: str | Path, market: str) -> ValidationResults:
#     """
#     Validates the local configuration files.
#
#     Args:
#         config_dir: Local path to the configuration files. Needs to follow a specific structure. See below.
#         market: The market to load the configuration for.
#
#     Configuration file structure:
#     ```
#     📦config_dir
#      ┣ 📂cogshop - The CogSHOP configuration
#      ┣ 📂market - The Market configuration for DayAhead, RKOM, and benchmarking.
#      ┣ 📂production - The physical assets configuration, Watercourse, PriceArea, Genertor, Plant  (SHOP centered)
#      ┗ 📜settings.yaml - Settings for resync.
#     ```
#     """
#


def plan(
    config_dir: Path,
    market: str,
    client: PowerOpsClient | None,
    model_names: str | list[str] | None = None,
    dump_folder: Optional[Path] = None,
) -> ModelDifferences:
    """
    Loads the local configuration files, transform them into Resync models, and compares them to the downloaded
    CDF Resync models.

    Args:
        config_dir: Local path to the configuration files. Needs to follow a specific structure. See below.
        market: The market to load the configuration for.
        client: The PowerOpsClient to use. If not provided, a new client will be created.
        model_names: The models to run the plan.
        dump_folder: If present, the local and CDF changes will be dumped to this directory. This is done so that
                    you can use local tools (for example, PyCharm or VS Code) to compare the detailed changes
                    between the local and CDF configuration.


    Returns:
        A ModelDifferences object containing the differences between the local and CDF configuration.

    Configuration file structure:
    ```
    📦config_dir
     ┣ 📂cogshop - The CogSHOP configuration
     ┣ 📂market - The Market configuration for DayAhead, RKOM, and benchmarking.
     ┣ 📂production - The physical assets configuration, Watercourse, PriceArea, Genertor, Plant  (SHOP centered)
     ┗ 📜settings.yaml - Settings for resync.
    ```
    """
    client = client or PowerOpsClient.from_settings()

    loaded_models = _load_transform(market, config_dir, client.cdf.config.project, model_names)

    logger.info(f"Load transform completed, models {', '.join([type(m).__name__ for m in loaded_models])} loaded")
    if client.datasets.read_dataset is None:
        raise ValueError("No read_dataset configured in settings")
    data_set_external_id = client.datasets.read_dataset
    all_differences = []
    for new_model in loaded_models:
        logger.info(f"Retrieving {new_model.model_name} from CDF")
        cdf_model = type(new_model).from_cdf(client, data_set_external_id=data_set_external_id)

        if isinstance(new_model, AssetModel):
            static_resources = new_model.static_resources_from_cdf(client)
        else:
            static_resources = {}

        differences = diff.model_difference(cdf_model, new_model, static_resources)
        _clean_relationships(client.cdf, differences, new_model)

        if dump_folder:
            dump_folder.mkdir(parents=True, exist_ok=True)
            # Standardize models for easy comparison
            new_model.standardize()
            cdf_model.standardize()

            (dump_folder / f"{new_model.model_name}_local.yaml").write_text(safe_dump(new_model.dump_as_cdf_resource()))
            (dump_folder / f"{new_model.model_name}_cdf.yaml").write_text(safe_dump(cdf_model.dump_as_cdf_resource()))

        all_differences.append(differences)
    return ModelDifferences(all_differences)


def apply(
    config_dir: Path,
    market: str,
    client: PowerOpsClient | None = None,
    model_names: str | list[str] | None = None,
    auto_yes: bool = False,
) -> ModelDifferences:
    """
    Loads the local configuration files, transform them into Resync models, and uploads it to CDF. Any deviations
    of the existing CDF configuration will be overwritten.

    Args:
        config_dir: Local path to the configuration files. Needs to follow a specific structure. See below.
        market: The market to load the configuration for.
        client: The PowerOpsClient to use. If not provided, a new client will be created.
        model_names: The models to run the plan.
        auto_yes: If true, all prompts will be auto confirmed.

    Returns:
        A ModelDifferences object containing the differences between the local and CDF configuration which have been
        written to CDF.

    Configuration file structure:
    ```
    📦config_dir
     ┣ 📂cogshop - The CogSHOP configuration
     ┣ 📂market - The Market configuration for DayAhead, RKOM, and benchmarking.
     ┣ 📂production - The physical assets configuration, Watercourse, PriceArea, Genertor, Plant  (SHOP centered)
     ┗ 📜settings.yaml - Settings for resync.
    ```
    """
    client = client or PowerOpsClient.from_settings()
    loaded_models = _load_transform(market, config_dir, client.cdf.config.project, model_names)

    written_changes = ModelDifferences([])
    for new_model in loaded_models:
        cdf_model = type(new_model).from_cdf(client, data_set_external_id=client.datasets.read_dataset)
        if isinstance(new_model, AssetModel):
            static_resources = new_model.static_resources_from_cdf(client)
        else:
            static_resources = {}

        differences = diff.model_difference(cdf_model, new_model, static_resources)

        # Do not create relationships to time series that does not exist.
        _clean_relationships(client.cdf, differences, new_model)
        # Remove the domain model as this is not CDF resources
        # TimeSeries are not updated by resync.
        differences.filter_out(group="Domain", field_names={"timeseries"})

        new_sequences_by_id = {s.external_id: s for s in new_model.sequences()}
        new_files_by_id = {f.external_id: f for f in new_model.files()}

        removed = _remove_resources(differences, client.cdf, auto_yes)
        added_updated = _add_update_resources(differences, client, auto_yes, new_sequences_by_id, new_files_by_id)

        written_changes.append(removed + added_updated)

    return written_changes


def destroy(
    client: PowerOpsClient | None,
    model_names: str | list[str] | None = None,
    auto_yes: bool = False,
    dry_run: bool = False,
) -> ModelDifferences:
    """
    Destroys all resync models in CDF. This will also delete all data in the models.

    Args:
        client: The PowerOpsClient to use. If not provided, a new client will be created.
        model_names: The models to destroy.
        auto_yes: If true, all prompts will be auto confirmed.
        dry_run: If true, the models will not be deleted, but the changes will be printed.

    Returns:
        A ModelDifferences object containing the resources that has been destroyed.
    """
    client = client or PowerOpsClient.from_settings()

    model_types = _to_models(model_names)
    destroyed = ModelDifferences([])
    for model_type in model_types:
        if issubclass(model_type, DataModel) and model_type.graph_ql:
            remove_data_model = _get_data_model_view_containers(
                client.cdf, model_type.graph_ql.id_, model_type.__name__
            )
            if not remove_data_model.changes:
                logger.warning(f"Skipping {model_type.__name__}, no data model found")
                continue
            static_resources = {}
        elif issubclass(model_type, DataModel) and model_type.dms_model:
            raise NotImplementedError()
        elif issubclass(model_type, DataModel) and model_type.source_model:
            raise NotImplementedError()
        elif issubclass(model_type, AssetModel):
            remove_data_model = ModelDifference(model_type.__name__, {})
            if issubclass(model_type, models.MarketModel):
                # We only need the root asset to be set.
                production_external_id = cast(str, models.ProductionModel.root_asset.external_id)
                model_type.set_root_asset("", "", "", production_external_id)
            static_resources = model_type.static_resources_from_cdf(client)
        else:
            raise ValueError(f"Unknown model type {model_type}")

        cdf_model = model_type.from_cdf(client, data_set_external_id=client.datasets.read_dataset)

        remove_data = diff.remove_only(cdf_model, static_resources)
        remove_data.filter_out(group="Domain", field_names={"timeseries"})

        if dry_run:
            destroyed.append(remove_data + remove_data_model)
        else:
            removed = _remove_resources(remove_data + remove_data_model, client.cdf, auto_yes)
            destroyed.append(removed)

    loader = DataModelLoader(Path.cwd() / "build")
    schema = loader.load()
    loader.destroy(client.cdf, schema, dry_run)

    # Spaces are deleted last, as they might contain other resources.
    spaces = set(space for d in model_types if issubclass(d, DataModel) for space in d.spaces())
    if spaces and not dry_run:
        deleted_space: list[SpaceId] = []
        # One at a time, in case there are other resources in the space that will prevent deletion.
        for space in spaces:
            logger.info(f"Deleting space {space}..")
            try:
                client.cdf.data_modeling.spaces.delete(list(spaces))
            except CogniteAPIError as e:
                logger.warning(f"Failed to delete space {space} with error {e}")
            else:
                logger.info(f"... deleted space {space}")
                deleted_space.append(SpaceId(space))
        if deleted_space:
            destroyed.append(
                ModelDifference(
                    model_name="All Models",
                    changes={
                        "spaces": FieldDifference(
                            group="CDF",
                            field_name="spaces",
                            removed=list(deleted_space),
                            added=[],
                            changed=[],
                            unchanged=[],
                        )
                    },
                )
            )
    if not dry_run:
        labels = AssetLabel.as_label_definitions() + RelationshipLabel.as_label_definitions()
        client.cdf.labels.delete([label.external_id for label in labels if label.external_id])

    return destroyed


def _remove_resources(differences: ModelDifference, cdf: CogniteClient, auto_yes: bool) -> ModelDifference:
    removed = ModelDifference(model_name=differences.model_name, changes={})
    for difference in sorted(differences, key=_remove_order):
        if not difference.removed:
            continue
        logger.info(f"Removals detected for {difference.field_name} in {differences.model_name}")
        logger.info(f"Remove count: {difference.as_summary().removed}")
        diff_ids = difference.as_ids(5)
        logger.info(f"Sample removed for {diff_ids.field_name}: {diff_ids.removed}")
        ans = "y" if auto_yes else input("Continue? (y/n)")
        if ans.lower() != "y":
            logger.info("Aborting")
            continue

        api = get_cognite_api(cdf, difference.field_name)

        api.delete(external_id=difference.removed_ids)

        logger.info(f"Deleted {len(difference.removed)} of {difference.field_name}")
        removed.changes[difference.field_name] = FieldDifference(
            group=difference.group,
            field_name=difference.field_name,
            added=[],
            removed=difference.removed,
            changed=[],
            unchanged=difference.unchanged,
        )
    return removed


def _add_update_resources(
    differences: ModelDifference,
    client: PowerOpsClient,
    auto_yes: bool,
    sequences_by_id: dict[str, CDFSequence],
    files_by_id: dict[str, CDFFile],
) -> ModelDifference:
    added_updated = ModelDifference(model_name=differences.model_name, changes={})
    for difference in sorted(differences, key=_adding_order):
        if not difference.added and not difference.changed:
            continue
        logger.info(f"Changes/Additions detected for {difference.field_name} in {differences.model_name}")
        summary_count = difference.as_summary()
        logger.info(f"Change count: {summary_count.changed}")
        logger.info(f"Addition count: {summary_count.added}")
        diff_ids = difference.as_ids(5)
        if diff_ids.changed:
            logger.info(f"Sample changed for {diff_ids.field_name}:")
            for change in difference.changed[:3]:
                logger.info(change.changed_fields)
        if diff_ids.added:
            logger.info(f"Sample added for {diff_ids.field_name}: {diff_ids.added}")
        ans = "y" if auto_yes else input("Continue? (y/n)")
        if ans.lower() != "y":
            logger.info("Aborting")
            continue

        difference.set_set_dataset(client.datasets.write_dataset_id)
        api = get_cognite_api(client.cdf, difference.field_name, sequences_by_id, files_by_id)

        if difference.added:
            api.create(difference.added)
            logger.info(f"Created {len(difference.added)} of {difference.field_name}")
            if difference.field_name in added_updated:
                added_updated[difference.field_name].added.extend(difference.added)
            else:
                added_updated[difference.field_name] = FieldDifference(
                    group=difference.group,
                    field_name=difference.field_name,
                    added=difference.added,
                    removed=[],
                    changed=[],
                    unchanged=[],
                )

        if difference.changed:
            updates = [c.new for c in difference.changed if not c.is_changed_content]
            if updates:
                api.upsert(updates, mode="replace")
                logger.info(f"Updated {len(updates)} of {difference.field_name}")
            content_updates = [c.new for c in difference.changed if c.is_changed_content]
            if content_updates:
                api.delete([c.external_id for c in content_updates if c.external_id])
                api.create(content_updates)
                logger.info(f"Updated {len(content_updates)} of {difference.field_name} with content")
            if difference.field_name in added_updated:
                added_updated[difference.field_name].changed.extend(difference.changed)
            else:
                added_updated[difference.field_name] = FieldDifference(
                    group=difference.group,
                    field_name=difference.field_name,
                    added=[],
                    removed=[],
                    changed=difference.changed,
                    unchanged=[],
                )
    return added_updated


def _remove_order(diff: FieldDifference) -> int:
    if diff.group == "Domain":
        return 0
    elif diff.group == "CDF" and diff.field_name not in {
        "edges",
        "nodes",
        "containers",
        "views",
        "data_models",
        "spaces",
        "parent_assets",
        "labels",
    }:
        return 1
    elif diff.group == "CDF" and diff.field_name == "edges":
        return 2
    elif diff.group == "CDF" and diff.field_name == "nodes":
        return 3
    elif diff.group == "CDF" and diff.field_name == "data_models":
        return 4
    elif diff.group == "CDF" and diff.field_name == "views":
        return 5
    elif diff.group == "CDF" and diff.field_name == "containers":
        return 6
    elif diff.group == "CDF" and diff.field_name == "spaces":
        return 7
    elif diff.group == "CDF" and diff.field_name == "parent_assets":
        return 8
    elif diff.group == "CDF" and diff.field_name == "labels":
        return 9
    else:
        return 8


def _adding_order(diff: FieldDifference) -> int:
    if diff.group == "Domain":
        return 0
    elif diff.group == "CDF" and diff.field_name in {"parent_assets", "labels"}:
        return 1
    elif diff.group == "CDF" and diff.field_name not in {"edges", "nodes"}:
        return 2
    elif diff.group == "CDF" and diff.field_name == "nodes":
        return 3
    elif diff.group == "CDF" and diff.field_name == "edges":
        return 4
    else:
        return 5


def _load_transform(market: str, path: Path, cdf_project: str, model_names: str | list[str] | None) -> list[Model]:
    model_types = _to_models(model_names)
    logger.info(f"Loading and transforming {', '.join([m.__name__ for m in model_types])}")

    config = ReSyncConfig.from_yamls(path, cdf_project)

    return transform(config, market, set(model_types))


def _to_models(model_names: str | list[str] | None) -> list[type[Model]]:
    if isinstance(model_names, str):
        model_names = [model_names]
    elif model_names is None:
        model_names = list(MODELS_BY_NAME)
    elif isinstance(model_names, list) and model_names and isinstance(model_names[0], str):
        model_names = model_names
    else:
        raise ValueError(f"Invalid model_names type: {type(model_names)}")

    if invalid := set(model_names) - set(MODELS_BY_NAME):
        raise ValueError(f"Invalid model names: {invalid}. Available models: {list(MODELS_BY_NAME)}")

    return [MODELS_BY_NAME[model_name] for model_name in model_names]


def _get_data_model_view_containers(cdf: CogniteClient, data_model_id: DataModelId, model_name: str) -> ModelDifference:
    data_model = cdf.data_modeling.data_models.retrieve(ids=data_model_id, inline_views=True)
    if not data_model:
        return ModelDifference(model_name=model_name, changes={})
    data_model = data_model.latest_version()
    views = data_model.views

    changes = {
        "views": FieldDifference(group="CDF", field_name="views", removed=list(ViewList(views).as_ids())),
        "containers": FieldDifference(
            group="CDF",
            field_name="containers",
            removed=list(
                {
                    prop.container
                    for view in views
                    for prop in view.properties.values()
                    if isinstance(prop, MappedProperty)
                }
            ),
        ),
        "data_models": FieldDifference(group="CDF", field_name="data_models", removed=[data_model.as_id()]),
    }

    return ModelDifference(model_name=model_name, changes=changes)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    TimeSeries,
)

if TYPE_CHECKING:
    from ._shop_price_scenario import SHOPPriceScenario, SHOPPriceScenarioGraphQL, SHOPPriceScenarioWrite


__all__ = [
    "SHOPPriceScenarioResult",
    "SHOPPriceScenarioResultWrite",
    "SHOPPriceScenarioResultApply",
    "SHOPPriceScenarioResultList",
    "SHOPPriceScenarioResultWriteList",
    "SHOPPriceScenarioResultApplyList",
    "SHOPPriceScenarioResultFields",
    "SHOPPriceScenarioResultTextFields",
]


SHOPPriceScenarioResultTextFields = Literal["price", "production"]
SHOPPriceScenarioResultFields = Literal["price", "production"]

_SHOPPRICESCENARIORESULT_PROPERTIES_BY_FIELD = {
    "price": "price",
    "production": "production",
}


class SHOPPriceScenarioResultGraphQL(GraphQLCore):
    """This represents the reading version of shop price scenario result, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop price scenario result.
        data_record: The data record of the shop price scenario result node.
        price: The price field.
        production: The production field.
        price_scenario: The price scenario field.
    """

    view_id = dm.ViewId("power-ops-day-ahead-bid", "SHOPPriceScenarioResult", "1")
    price: Union[TimeSeries, dict, None] = None
    production: Union[TimeSeries, dict, None] = None
    price_scenario: Optional[SHOPPriceScenarioGraphQL] = Field(None, repr=False, alias="priceScenario")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("price_scenario", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> SHOPPriceScenarioResult:
        """Convert this GraphQL format of shop price scenario result to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return SHOPPriceScenarioResult(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            price=self.price,
            production=self.production,
            price_scenario=(
                self.price_scenario.as_read() if isinstance(self.price_scenario, GraphQLCore) else self.price_scenario
            ),
        )

    def as_write(self) -> SHOPPriceScenarioResultWrite:
        """Convert this GraphQL format of shop price scenario result to the writing format."""
        return SHOPPriceScenarioResultWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            price=self.price,
            production=self.production,
            price_scenario=(
                self.price_scenario.as_write() if isinstance(self.price_scenario, DomainModel) else self.price_scenario
            ),
        )


class SHOPPriceScenarioResult(DomainModel):
    """This represents the reading version of shop price scenario result.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop price scenario result.
        data_record: The data record of the shop price scenario result node.
        price: The price field.
        production: The production field.
        price_scenario: The price scenario field.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power-ops-types", "SHOPPriceScenarioResult"
    )
    price: Union[TimeSeries, str, None] = None
    production: Union[TimeSeries, str, None] = None
    price_scenario: Union[SHOPPriceScenario, str, dm.NodeId, None] = Field(None, repr=False, alias="priceScenario")

    def as_write(self) -> SHOPPriceScenarioResultWrite:
        """Convert this read version of shop price scenario result to the writing version."""
        return SHOPPriceScenarioResultWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            price=self.price,
            production=self.production,
            price_scenario=(
                self.price_scenario.as_write() if isinstance(self.price_scenario, DomainModel) else self.price_scenario
            ),
        )

    def as_apply(self) -> SHOPPriceScenarioResultWrite:
        """Convert this read version of shop price scenario result to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class SHOPPriceScenarioResultWrite(DomainModelWrite):
    """This represents the writing version of shop price scenario result.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop price scenario result.
        data_record: The data record of the shop price scenario result node.
        price: The price field.
        production: The production field.
        price_scenario: The price scenario field.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power-ops-types", "SHOPPriceScenarioResult"
    )
    price: Union[TimeSeries, str, None] = None
    production: Union[TimeSeries, str, None] = None
    price_scenario: Union[SHOPPriceScenarioWrite, str, dm.NodeId, None] = Field(None, repr=False, alias="priceScenario")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            SHOPPriceScenarioResult, dm.ViewId("power-ops-day-ahead-bid", "SHOPPriceScenarioResult", "1")
        )

        properties: dict[str, Any] = {}

        if self.price is not None or write_none:
            if isinstance(self.price, str) or self.price is None:
                properties["price"] = self.price
            else:
                properties["price"] = self.price.external_id

        if self.production is not None or write_none:
            if isinstance(self.production, str) or self.production is None:
                properties["production"] = self.production
            else:
                properties["production"] = self.production.external_id

        if self.price_scenario is not None:
            properties["priceScenario"] = {
                "space": self.space if isinstance(self.price_scenario, str) else self.price_scenario.space,
                "externalId": (
                    self.price_scenario if isinstance(self.price_scenario, str) else self.price_scenario.external_id
                ),
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        if isinstance(self.price_scenario, DomainModelWrite):
            other_resources = self.price_scenario._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.price, CogniteTimeSeries):
            resources.time_series.append(self.price)

        if isinstance(self.production, CogniteTimeSeries):
            resources.time_series.append(self.production)

        return resources


class SHOPPriceScenarioResultApply(SHOPPriceScenarioResultWrite):
    def __new__(cls, *args, **kwargs) -> SHOPPriceScenarioResultApply:
        warnings.warn(
            "SHOPPriceScenarioResultApply is deprecated and will be removed in v1.0. Use SHOPPriceScenarioResultWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "SHOPPriceScenarioResult.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class SHOPPriceScenarioResultList(DomainModelList[SHOPPriceScenarioResult]):
    """List of shop price scenario results in the read version."""

    _INSTANCE = SHOPPriceScenarioResult

    def as_write(self) -> SHOPPriceScenarioResultWriteList:
        """Convert these read versions of shop price scenario result to the writing versions."""
        return SHOPPriceScenarioResultWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> SHOPPriceScenarioResultWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class SHOPPriceScenarioResultWriteList(DomainModelWriteList[SHOPPriceScenarioResultWrite]):
    """List of shop price scenario results in the writing version."""

    _INSTANCE = SHOPPriceScenarioResultWrite


class SHOPPriceScenarioResultApplyList(SHOPPriceScenarioResultWriteList): ...


def _create_shop_price_scenario_result_filter(
    view_id: dm.ViewId,
    price_scenario: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if price_scenario and isinstance(price_scenario, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("priceScenario"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": price_scenario},
            )
        )
    if price_scenario and isinstance(price_scenario, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("priceScenario"),
                value={"space": price_scenario[0], "externalId": price_scenario[1]},
            )
        )
    if price_scenario and isinstance(price_scenario, list) and isinstance(price_scenario[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("priceScenario"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in price_scenario],
            )
        )
    if price_scenario and isinstance(price_scenario, list) and isinstance(price_scenario[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("priceScenario"),
                values=[{"space": item[0], "externalId": item[1]} for item in price_scenario],
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from .core import DATAMODEL_ID_TO_RESYNC_NAME, MODELS_BY_NAME, apply, destroy, init, plan, validate
from .core.echo import Echo
from .v1 import apply as apply_v1
from .v1 import plan as plan_v1

__all__ = [
    "plan_v1",
    "apply_v1",
    "apply",
    "plan",
    "init",
    "destroy",
    "validate",
    "MODELS_BY_NAME",
    "Echo",
    "DATAMODEL_ID_TO_RESYNC_NAME",
]
